/*
 * ExpDAG.h: interface for the ExpDAG class.
 */

#if !defined(AFX_BOOLEANDAG_H__61A0B2EB_5CC4_4D69_AA64_9DC156ED4C8D__INCLUDED_)
#define AFX_BOOLEANDAG_H__61A0B2EB_5CC4_4D69_AA64_9DC156ED4C8D__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include <vector>
#include <string>
#include <list>
#include <map>
#include <set>
#include <stack>

#include "ExpressionNodes.h"
#include "BasicError.h"
#include "timerclass.h"
#include "NodeVisitor.h"
#include "Dllist.h"

#ifdef CONST
#undef CONST
#endif

using namespace std;
// #define SCHECKMEM

inline bool comp_id(ExpNode* n1, ExpNode* n2){
  int n1id = n1->globalId;
  if(n1id == -1){
    return false;
  }
  int n2id = n2->globalId;
  if(n2id == -1){ //& n1->id != -1
    return n1id!= n2id;
  }
  return n1id < n2id;
}

// NOTE(JY): What does this do?
extern timerclass TTMMPP;

/**
 * ExpDag class for expression DAGs.
 */
class ExpDAG  
{
#ifdef SCHECKMEM
	static set<ExpDAG*> allocated;
#endif

  int offset;
  int intSize;
  bool ownsNodes;

  vector<ExpNode*> nodes;

  const string name;

  void shareparent_remove(int i);

  int n_inputs;

public:
	typedef vector<ExpNode*>::iterator iterator;
	typedef vector<ExpNode*>::reverse_iterator reverse_iterator;

	ExpDAG(const string& name_="anon");
	virtual ~ExpDAG();

  /**
   * Mutators for graph creation.
   */
  int getIntSize(){
	return intSize;
  }

  /**
	Creates an N-bit output named gen_name, connected to internal node nodeToOutput.
  */
  ExpNode* new_node(ExpNode* mother, ExpNode* father, ExpNode::Type t);

  /**
  * Mutators for graph cleanup and refactoring.
  */
  void removeNullNodes();
  void replace(int original, ExpNode* replacement);

  void addNewNode(ExpNode* v);
  void addNewNodes(vector<ExpNode*>& v);
  void clearBackPointers();

  /*
  void change_father(const vname& father, const vname& son);
  void change_mother(const vname& father, const vname& son);
  */

  void layer_graph();  
  void cleanup(); //Sorts and cleans up the graph.
  void cleanUnshared();
  void clear();
  // void rename(const vname& oldname,  const vname& newname);
  void resetBackPointers();

  void setOffset(int ofs){ offset = ofs; }
  // ExpNode* create_inter(
  //    int n, const vname& gen_name, int&counter, ExpNode::Type type);
  // ExpNode* create_inputs(int n, const vname& gen_name);
  void printSlice(ExpNode* node, ostream& out)const;
  /*Both of these functions destroy their argument*/
  // virtual void makeMiter(ExpDAG* bdag);
  // virtual void andDag(ExpDAG* bdag);
  
  /**
   * Field Access methods.	
   */
  /*
  ExpNode* const & operator[](int idx)const{ return nodes[idx]; };
  ExpNode*& operator[](int idx){ return nodes[idx]; };
  int size()const {return nodes.size();}
  
  // ExpNode* get_node(const vname& name);
  // ExpNode* unchecked_get_node(const vname& name);
  iterator begin(){ return nodes.begin(); }
  iterator end(){ return nodes.end(); }

  reverse_iterator rbegin(){ return nodes.rbegin(); }
  reverse_iterator rend(){ return nodes.rend(); }

  inline bool checkNodePosition(ExpNode* bn){
	  int bid = bn->globalId;
	return bid < this->size() && bid >= 0 && bn == nodes[bid];
  }
  */
  const string& get_name()const{
	  return name;
  }
  
  void sliceH(ExpNode* n, ExpDAG* bd);
  
  void repOK();

  ExpDAG* clone();
  void clone_nodes(vector<ExpNode*>& nstore, Dllist* dl=NULL);

  void print(ostream& out)const;
  void lprint(ostream& out);

};

#endif // !defined(AFX_BOOLEANDAG_H__61A0B2EB_5CC4_4D69_AA64_9DC156ED4C8D__INCLUDED_)

