#include "ExpDAGCreator.h"

ExpDAGCreator::ExpDAGCreator(ExpDAG* p_dag):dag(p_dag),optim(*p_dag) { }

ExpDAGCreator::~ExpDAGCreator(void) { }

ExpNode* ExpDAGCreator::simplify(ExpNode* node) {
	return optim.simplify(node);
}

ExpNode* ExpDAGCreator::optimizeAndAdd(ExpNode* node){
	ExpNode* tmp;

	tmp = optim.computeOptim(node);

	Assert(tmp->globalId != -22, "optimized node should not have been garbage collected");
	if(tmp == node) {
#ifdef __DEBUG_1
		cout << "ExpDAGCreator::optimizeAndAdd: Adding node " << node->lprint()
			 << endl;
		cout << "NODE " << node->globalId << " WAS NOT OPTIMIZED." << endl;
		cout << endl << node->lprint() << endl;
#endif
		optim.addNode(node);
	} else{
#ifdef __DEBUG_1
		cout << "ExpDAGCreator::optimizeAndAdd: Remove node " << node->globalId
			 << endl;
		cout << "ExpDAGCreator::optimizeAndAdd: Replacing it with "
			 << tmp->globalId << ": "
			 << tmp->lprint() << endl;
#endif
		optim.remove_node(node);
	}

#ifdef __TEST
	Assert(tmp != NULL, "new node should not be null");
	Assert(tmp->globalId != -22, "new node should not have been garbage collected");
#endif
	return tmp;
}

ExpNode* ExpDAGCreator::new_node(ExpNode* mother, 
                                ExpNode* father,  ExpNode* thenode){
  thenode->set_parents(mother, father);

  int s = thenode->get_name();
  bool sym = thenode->sym();
  thenode = optimizeAndAdd(thenode);
#ifdef __DEBUG_1
  Assert(thenode != NULL, "Node should not be null");
  Assert(thenode->globalId != -22, "Node should not be garbage in new_node");
#endif
  thenode->set_sym(sym);
  if(s > -1) {
	  dag->addNewNode(thenode);
  }

#ifdef __DEBUG_1
  Assert(thenode->globalId != -22, "Node has not been garbage collected");
  cout << "ExpDagCreator::new_node: Created node with EID "
	   << thenode->get_name()
	   << "= " << thenode->lprint() << endl;
#endif
  return thenode;
}


ExpNode* ExpDAGCreator::create_const(int n){
	ConstNode* c = optim.getCnode(n);
	dag->setOffset(optim.newNodesSize());	
	return c;
}
