/*
 * What is this file used for?
 */
#pragma once

#include "BackendTypes.h"
#include "ExpDAG.h"
#include "DagOptim.h"

class ExpDAGCreator
{
	ExpDAG* const dag;
	DagOptim optim;

public:
	map<int, vector<ExpNode*> > methdparams;
	ExpDAGCreator(ExpDAG* p_dag);
	virtual ~ExpDAGCreator(void);
  virtual ExpNode* create_const(int n);

  /* TODO: Do we still need inputs and outputs? */
  ExpNode* create_inputs(int n, const string& gen_name=string("INPUT"));

  /**
	Creates an N-bit output named gen_name connected to internal node nodeToOutput.
  */
  ExpNode* create_outputs(
      int n, ExpNode* nodeToOutput, const string& gen_name=string("OUTPUT"));
  ExpNode* create_outputs(int n, const string& gen_name=string("OUTPUT"));
  

  /** Produces a node called name.
  */
  ExpNode* get_node(const vname& name);

  /* Used during initialization when creating ARITH nodes and ASSERT nodes.
   * thenode will always be an ARITH node or an ASSERT node.	The node thenode
   * should not be used after calls to this function, because this function may
   * optimize it away and deallocate it; so one must be very very careful to
   * always use the return value instead.  Also, all the fields of the arith
   * node should already be initialized. Otherwise, you'll get wrong behavior.
	*/
  ExpNode* new_node(ExpNode* mother, 
                      ExpNode* father, ExpNode* thenode);
  ExpNode* simplify(ExpNode* node);

  /*-------------------------------------*/

  ExpNode* optimizeAndAdd(ExpNode* node);

  void remove_node(ExpNode* node) { optim.remove_node(node); }
  /*
  ExpNode* get_alias(const vname& s){
    return get_node(s);
  }
  */

 void finalize(){
	 optim.cleanup(*dag);
 }

};
