// ExpDAG.cpp: implementation of the ExpDAG class.
//
//////////////////////////////////////////////////////////////////////

#include "ExpDAG.h"
#include "BasicError.h"
#include "SATSolver.h"

#include <sstream>
#include <algorithm>


using namespace std;

/**
 * Construction/Destruction
 */
#ifdef SCHECKMEM
set<ExpDAG*> ExpDAG::allocated;
#endif 

ExpDAG::ExpDAG(const string& name_):name(name_)
{
  has_passthrough=false;
  is_layered=false;
  is_sorted=false;
  offset = 0;
  ownsNodes = true;
  n_inputs = 0;
  intSize = 2;

#ifdef SCHECKMEM
  cout<<"Checking allocation"<<endl;
  allocated.insert(this);
#endif
}

void ExpDAG::sliceH(ExpNode* n, ExpDAG* bd){
	if(n->flag == 0){
		n->flag = 1;
		if(n->get_mother() != NULL){
			sliceH(n->get_mother(), bd);
		}
		if(n->get_father() != NULL){
			sliceH(n->get_father(), bd);
		}
		bd->addNewNode(n);
	}
}
/*
ExpDAG* ExpDAG::slice(int i, AssertNode*& out){
	for(ExpDAG::iterator it = nodes.begin(); it != nodes.end(); ++it){
		(*it)->flag = 0;
	}
	ExpDAG* bd = new ExpDAG(this->name);
	bd->ownsNodes = false;
	bd->sliceH(nodes[i], bd);
	if(out->mother != NULL){
		out->mother->mother = nodes[i];		
		if(out->mother->father != NULL){
			bd->addNewNode(out->mother->father);
		}
		bd->addNewNode(out->mother);
	}else{
		out->mother = nodes[i];
	}
	bd->addNewNode(out);
	return bd;
}
*/


void ExpDAG::clear(){	
	if(ownsNodes){
	  for(int i=0; i < nodes.size(); ++i){
  		delete nodes[i];
  		nodes[i] = NULL;
	  }
	}
  nodes.clear();
  __cur_nodes.clear();
}


ExpDAG::~ExpDAG()
{
#ifdef SCHECKMEM
	allocated.erase(this);
#endif
}

void ExpDAG::replace(int original, ExpNode* replacement){	
	int i = original;
	Assert( i < nodes.size() && i >= 0, "Out of bounds violation "<<i<<" >= "<<nodes.size()<<endl);
	Assert( replacement != NULL, "Why are you replacing with a null replacement");
	ExpNode* onode = nodes[i];
	Assert( onode != NULL, "This can't happen");
	
	onode->neighbor_replace(replacement);
	
	nodes[i]->globalId = -22;
	delete nodes[i];
  	nodes[i] = NULL;
}


void ExpDAG::removeNullNodes(){
	int nullnodes = 0;
	for(vector<ExpNode*>::iterator it = nodes.begin(); it != nodes.end(); ++it){
		if(*it != NULL) nullnodes++;
	} 
	if( nullnodes < nodes.size()){
		vector<ExpNode*> newnodes(nullnodes);
		nullnodes = 0;
		for(vector<ExpNode*>::iterator it = nodes.begin(); it != nodes.end(); ++it){
			if(*it != NULL){
				newnodes[nullnodes] = *it;
				nullnodes++;
			}
		}
		swap(newnodes, nodes);
		Assert(nodes.size() == nullnodes, "If this fails, it means I don't know how to use STL");
		Dout( cout<<"Removing "<<newnodes.size() - nodes.size()<<" nodes"<<endl );
	}
}


void ExpDAG::shareparent_remove(int i){
	
  ExpNode* onode = nodes[i];	
	
  Assert( onode->get_father() == onode->get_mother()
		, "This must be true, otherwise, the compiler is wrong");
  Assert( onode->get_father() != NULL, "Can this happen? To me? Nah  ");
 
  onode->globalId = -22;
  delete onode;
  nodes[i] = NULL;  
}



void ExpDAG::repOK(){

	/*
	map<ExpNode::Type, set<ExpNode*> > tsets;
	for(map<ExpNode::Type, vector<ExpNode*> >::iterator it = nodesByType.begin(); it != nodesByType.end(); ++it){
		ExpNode::Type type = it->first;
		vector<ExpNode*>& nv = it->second;
		for(int i=0; i<nv.size(); ++i){
			Assert(tsets[type].count(nv[i])==0, "You have a repeated node in the list");
			tsets[type].insert(nv[i]);
			Assert(nv[i]->type >= 0, "This is corrupted!!");
		}		
	}

	cout<<"*** DOING REPOK ****"<<endl;
	//First, we check that the array doesn't contain any repeated nodes.
	map<ExpNode*, int> nodeset;
	for(int i=0; i<nodes.size(); ++i){
		if(nodes[i] != NULL){
		Assert(nodeset.count(nodes[i]) == 0, "There is a repeated node!!! it's in pos "<<i<<" and "<<nodeset[nodes[i]] );		
		nodeset[nodes[i]] = i;
		}
		if(tsets.count(nodes[i]->type)>0){
			Assert( tsets[nodes[i]->type].count(nodes[i])==1 , "All the typed nodes should be accounted for in the nodesByType");
		}
	}
	*/
	DllistNode* cur = this->assertions.head;
	DllistNode* last=NULL;
	Assert(last == this->assertions.tail, "Missing nodes" );
}


void ExpDAG::cleanUnshared(){
  for(int i=0; i < nodes.size(); ++i){
    ExpNode* onode = nodes[i];
    if(onode->flag == 0 ){
      onode->globalId = -22;
      delete onode;
      nodes[i] = NULL;
    }
  }
  removeNullNodes();  
}



//This routine removes nodes that do not flow to the output.
void ExpDAG::cleanup(){
  //The first optimization is to remove nodes that don't contribute to the output. 	  
  for(int i=0; i < nodes.size(); ++i){
    nodes[i]->flag = 0;
  }
  int idx=0;

  for(int i=0; i < nodes.size(); ++i){
    if(nodes[i]->flag == 0 &&
       nodes[i]->type != ExpNode::SYM){
      //nodes[i]->id = idx;
      ++idx;
    }
  }
  for(int i=0; i < nodes.size(); ++i){
    ExpNode* onode = nodes[i];
    if(onode->flag == 0 &&
        onode->type != ExpNode::SYM) {

      onode->globalId = -22;
      delete onode;
      nodes[i] = NULL;
    }
  }  
  removeNullNodes();  
  sort(nodes.begin(), nodes.end(), comp_id);
}

/*
void ExpDAG::change_father(const vname& father, const vname& son){
  try{
    Assert(__named_nodes.find(father) != __named_nodes.end(), "The father name does not exist: "<<father);  
  }catch(BasicError& be){
    throw be;
  }
  Assert(__named_nodes.find(son) != __named_nodes.end(), "The son name does not exist.");  
  Assert(__named_nodes[son]->get_father() == NULL, "You should not call this function if you already have a father.");
  ExpNode* cur = __named_nodes[son]->get_father();
  cur = __named_nodes[father];
}

void ExpDAG::change_mother(const vname& mother, const vname& son){
  try{
    Assert(__named_nodes.find(mother) != __named_nodes.end(), "The mother name does not exist: "<<mother);
  }catch(BasicError& be){    
    throw be;
  }
  Assert(__named_nodes.find(son) != __named_nodes.end(), "The son name does not exist.");  
  Assert(__named_nodes[son]->get_mother() == NULL, "You should not call this function if you already have a father.");
  ExpNode* cur = __named_nodes[son]->get_mother();
  cur = __named_nodes[mother];
}
*/

void ExpDAG::addNewNode(ExpNode* node){
	Assert( node != NULL, "null node can't be added.");
	Assert( node->globalId != -22, "This node should not exist anymore");
	nodes.push_back(node);
	/*
	if((node->get_name()) > -1){
		__cur_nodes.push(node);
	}
	*/
}

void ExpDAG::addNewNodes(vector<ExpNode*>& v){
	//Assume all the nodes in v are already part of the network, meaning all their parents and children are properly set.
	nodes.reserve(nodes.size() + v.size());
	for(int i=0; i<v.size(); ++i){
		if(v[i] != NULL){
			addNewNode(v[i]);
		}
	}
}


ExpNode* ExpDAG::unchecked_get_node(const vname& name){
  ExpNode* fth;
  if(name == -1){
    fth = NULL;
  }else{
	  if(name <= __cur_top) {
		fth = __cur_nodes[name];
	  }else{
		fth = NULL;
	  }
  }
  return fth;
}

/*
ExpNode* ExpDAG::get_node(const vname& name){
  ExpNode* fth;
  //Assert(name.size()==0 || named_nodes.find(name) != named_nodes.end(), "name does not exist: "<<name);
  if(name == -1){
    fth = NULL;
  }else{
	if(name <= __cur_top) {
		fth = __cur_nodes[name];
	}else{
		cout << "ExpDAG::get_node: " << name << " was not found." << endl;
		Assert(false, "Node should be in list");
	}
  }
  return fth;
}
*/

ExpNode* ExpDAG::new_node(ExpNode* mother, 
                          ExpNode* father, ExpNode::Type t){
                                	
  ExpNode* tmp = newBoolNode(t);
  tmp->set_parents(mother, father);
  Assert( tmp->globalId != -22, "This node should not exist anymore");
  nodes.push_back(tmp);  
  return tmp;
}

ExpNode* ExpDAG::create_inter(
    int n, const vname& gen_name, int& counter,  ExpNode::Type type){
  //Create interface nodes, either source, dest, or ctrl.

  if(gen_name <= __cur_top) {
    return __cur_nodes[gen_name];
  }

  if(n < 0){
    ExpNode* tmp = newBoolNode(type);
    // nodesByType[type].push_back(tmp);
    nodes.push_back(tmp);
    // tmp->set_name(gen_name);
    __cur_nodes[gen_name] = tmp;
    // __cur_top = max(__cur_top, gen_name);
    tmp->otype = ExpNode::BOOL;
    ++counter;
  return tmp;
  }else{
    ExpNode* tmp = newBoolNode(type);
    // nodesByType[type].push_back(tmp);
    dynamic_cast<InterNode*>(tmp)->set_nbits(n);
    nodes.push_back(tmp);
    //tmp->set_name(gen_name);
    __cur_nodes[gen_name] = tmp;
    tmp->otype = ExpNode::INT;
    counter += n;
  return tmp;
  }
}


ExpNode* ExpDAG::create_inputs(int n, const vname& gen_name){
  return create_inter(n, gen_name, n_inputs, ExpNode::SYM);
}

void ExpDAG::printSlice(ExpNode* node, ostream& out)const{    
	out<<"digraph G{"<<endl;
	set<const ExpNode* > s;

	node->printSubDAG(out, s);
	

	out<<"}"<<endl;
}



void ExpDAG::print(ostream& out)const{    
  out<<"digraph G{"<<endl;
  for(int i=0; i<nodes.size(); ++i){
  	if(nodes[i] != NULL){
  		nodes[i]->outDagEntry(out);
  	}    
  }

  out<<"}"<<endl;
}

void ExpDAG::lprint(ostream& out){    
  out<<"dag{"<<endl;
  for(int i=0; i<nodes.size(); ++i){
  	if(nodes[i] != NULL){
  		out<<nodes[i]->lprint()<<endl;
  	}    
  }

  out<<"}"<<endl;
}


void ExpDAG::clearBackPointers(){
  // This used to clear pointers in children.
}

void ExpDAG::resetBackPointers(){

}



void ExpDAG::andDag(ExpDAG* bdag){
  ExpNode* tip = NULL;
  map<ExpNode*, ExpNode*> replacements;

  for(ExpDAG::iterator node_it = bdag->begin(); node_it != bdag->end(); ++node_it){
    Assert( (*node_it) != NULL, "Can't make a miter when you have null nodes.");
    (*node_it)->flag = 0;
    Dout( cout<<" adding "<<(*node_it)->get_name()<<endl );

    if( (*node_it)->type != ExpNode::SYM){
      nodes.push_back( (*node_it) );
      (*node_it)->switchInputs(*this, replacements);
      if((*node_it)->type == ExpNode::ASSERT ){
        // nodesByType[(*node_it)->type].push_back((*node_it));
        DllistNode* tt = getDllnode((*node_it));
        tt->remove();
        assertions.append( tt );
      }
    }

    Assert(false, "Didn't know we get here.");
    /*
    if( (*node_it)->type == ExpNode::SYM ){
      if( !has_name((*node_it)->get_name()) ){
        nodes.push_back( (*node_it) );
        // nodesByType[(*node_it)->type].push_back((*node_it));
        // TODO: Is this what to do?
        __cur_nodes[__cur_top++] = *node_it;
      }else{
        replacements[*node_it] = this->get_node((*node_it)->get_name());
        delete (*node_it);
      }
    }
    */

  }

  removeNullNodes();
  cleanup();
  delete bdag;
}


void ExpDAG::makeMiter(ExpDAG* bdag){
  ExpNode* tip = NULL; 
  map<ExpNode*, ExpNode*> replacements;	

  for(ExpDAG::iterator node_it = bdag->begin(); node_it != bdag->end(); ++node_it){
    Assert( (*node_it) != NULL, "Can't make a miter when you have null nodes.");
    (*node_it)->flag = 0;
    Dout( cout<<" adding "<<(*node_it)->get_name()<<endl );

    if( (*node_it)->type != ExpNode::SYM) {
      nodes.push_back( (*node_it) );
      (*node_it)->switchInputs(*this, replacements);
      if( (*node_it)->type == ExpNode::ASSERT ){
        //nodesByType[(*node_it)->type].push_back((*node_it));
        DllistNode* tt = getDllnode((*node_it));
        tt->remove();
        assertions.append( tt );
      }
    }

    Assert(false, "Didn't know we get here.");
    /*
    if( (*node_it)->type == ExpNode::SYM ){
      if( !has_name((*node_it)->get_name()) ){
        nodes.push_back( (*node_it) );
        //nodesByType[(*node_it)->type].push_back((*node_it));
        //__named_nodes[(*node_it)->get_name()] = *node_it;
      }else{
        replacements[*node_it] = this->get_node((*node_it)->get_name());
        delete (*node_it);
      }
    }
    */
  }

  removeNullNodes();
  cleanup();
  delete bdag;
}

/*
void ExpDAG::rename(const vname& oldname,  const vname& newname){
  ExpNode* node = __named_nodes[oldname];
  //node->set_name(newname);
  __named_nodes.erase(oldname);
  __named_nodes[newname] = node;
}
*/


void ExpDAG::clone_nodes(vector<ExpNode*>& nstore, Dllist* dl){
  nstore.resize(nodes.size());

  Dout( cout<<" after relabel "<<endl );
  int nnodes = 0;
  for(ExpDAG::iterator node_it = begin(); node_it != end(); ++node_it){
    if( (*node_it) != NULL ){		
      Assert( (*node_it)->globalId != -22 , "This node has already been deleted "<<	(*node_it)->get_name() );
      ExpNode* bn = (*node_it)->clone(false);

			if( dl != NULL && isDllnode(bn) ){
				dl->append(getDllnode(bn));
			}

			Dout( cout<<" node "<<(*node_it)->get_name()<<" clones into "<<bn->get_name()<<endl );
			nstore[nnodes] = bn;
			nnodes++;
		}else{
			Assert( false, "The graph you are cloning should not have any null nodes.");
		}
	}
	Dout( cout<<" after indiv clone "<<endl );
	nstore.resize(nnodes);
	ExpDAG::iterator old_it = begin();
  /*
	for(ExpDAG::iterator node_it = nstore.begin(); node_it != nstore.end(); ++node_it, ++old_it){
		(*node_it)->redirectPointers(*this, (vector<const ExpNode*>&) nstore, (*old_it)->children);
	}
  */
}



ExpDAG* ExpDAG::clone(){
	Dout( cout<<" begin clone "<<endl );

	//clone_nodes(bdag->nodes, &bdag->assertions);

	Dout( cout<<" after redirectPointers "<<endl );
	//bdag->intSize = intSize;

	/*
	for(map<eid_t, ExpNode*>::iterator it = __named_nodes.begin();
      it != __named_nodes.end(); ++it){
		Assert( it->second->globalId != -22 , "This node has already been deleted "<<it->first<<endl );
		Assert( bdag->nodes.size() > it->second->globalId, " Bad node  "<<it->first<<endl );
		(bdag->__named_nodes[it->first]) = bdag->nodes[it->second->globalId];
	}
	*/
	
	/*
	for(map<ExpNode::Type, vector<ExpNode*> >::iterator it =nodesByType.begin(); 
					it != nodesByType.end(); ++it){
		vector<ExpNode*>& tmp = bdag->nodesByType[it->first];
		Assert( tmp.size() == 0, "This can't happen. This is an invariant.");
		for(int i=0; i<it->second.size(); ++i){
			Assert( it->second[i]->id != -22 , "This node has already been deleted "<<it->second[i]->get_name()<<endl );			
			tmp.push_back( bdag->nodes[ it->second[i]->id ] );	
		}							
	}
	*/
	//return bdag;
}
