#include "ExpressionNodes.h"
#include "ExpDAG.h"

#include <fstream>
#include <sstream>
#include <algorithm>
#include <string>

/**
 * ConstNode
 */
ConstNode::ConstNode() : ExpNode() {
	type = CONST;
	__val = -1;
	otype = INT;
  ytype = INT;
}
ConstNode::ConstNode(int n) : ExpNode() {
	type = CONST;
	__val = n;
	otype = INT;
  ytype = INT;
}
ConstNode::ConstNode(const ConstNode& bn) :
	ExpNode(bn), __val(bn.__val) {
	//if(val == 13){ cout<<" surprise"<<endl; }
}

string ConstNode::get_namestr() const{
  stringstream str;
  str<< globalId <<"_C";
  if( __val<0){
    str<<"m";
  }
  str<<abs(__val);
  return str.str();
}

yices_expr ConstNode::to_yices(yices_context yctxt) {
  if (ytype == ExpNode::BOOL) {
    if (__val) {
      return yices_mk_true(yctxt);
    } else {
      return yices_mk_false(yctxt);
    }
  } else {
    return yices_mk_num(yctxt, __val);
  }
}

string ConstNode::get_yicesstr() const{
  stringstream str;
  if (ytype == ExpNode::BOOL) {
	  if (__val == 1) { str << "true"; }
	  else { str << "false"; }
  } else {
	  str << __val;
  }
  return str.str();
}

string ConstNode::lid(){
  stringstream str;
  str<<"(" << __val << ")";
  return str.str();
}
string ConstNode::lprint(){
  stringstream str;
  str<<"(" << __val << ")";
  return str.str();
}
ExpNode* ConstNode::clone(bool copyChildren) {
  return new ConstNode(*this);
};

EnvNode ConstNode::eqNode(ExpNode* other) {
	if (other->getType() == ExpNode::CONST) {
		return (__val == ((ConstNode*)other)->getVal());
	} else if (ExpNode::SYM == other->getType()) {
    EqNode* en = new EqNode();
    en->set_parents(this, other);
		return EnvNode(en);
	} else {
    return EnvNode(false);
  }
}

/**
 * StringNode
 */
StringNode::StringNode() : ExpNode() {
  type = STRING;
  otype = STRINGTY;
  ytype = BOTTOM;
}
StringNode::StringNode(string n) : ExpNode() {
  type = CONST;
  __val = n;
  otype = INT;
  ytype = INT;
}
StringNode::StringNode(const StringNode& bn) :
  ExpNode(bn), __val(bn.__val) {
}

string StringNode::get_namestr() const{
  stringstream str;
  str<< globalId <<"_S";
  str<<__val;
  return str.str();
}

yices_expr StringNode::to_yices(yices_context yctxt) {
  Assert(false, "Should not be putting a string into Yices.");
}

string StringNode::get_yicesstr() const{
  Assert(false, "Should not be putting a string into Yices.");
}

string StringNode::lid(){
  stringstream str;
  str<<"(" << __val << ")";
  return str.str();
}
string StringNode::lprint(){
  stringstream str;
  str<<"(" << __val << ")";
  return str.str();
}
ExpNode* StringNode::clone(bool copyChildren) {
  return new StringNode(*this);
};

EnvNode StringNode::eqNode(ExpNode* other) {
  if (other->getType() == ExpNode::STRING) {
    const string& s = ((StringNode*)other)->getVal();
    return EnvNode(__val.compare(s) == 0);
  } else {
    return EnvNode(false);
  }
}

/**
 * RecordNode
 */
RecordNode::RecordNode() : ExpNode() {
  type = RECORD;
  /* TODO */
  __record = new vector<EnvNode>;
  __num_fields = 0;
  __sym = false;
}
RecordNode::~RecordNode() {
  delete __record;
}
RecordNode::RecordNode(size_t numfields) : ExpNode() {
#ifdef __DEBUG_1
	cout << "RecordNode::RecordNode: Creating record with " << numfields
		 << " fields." << endl;
#endif
	type = RECORD;
	__record = new vector<EnvNode>(numfields, EnvNode(0));
	__num_fields = numfields;
	__sym = false;
}
RecordNode::RecordNode(const RecordNode& bn)
  : ExpNode(bn), __record(bn.__record), __num_fields(bn.__num_fields) { }

void RecordNode::accept(NodeVisitor& visitor)  { visitor.visit( *this ); }
string RecordNode::get_namestr() const{
  stringstream str;
  str<< globalId <<"_C";
/*  str<<abs(val); */
  return str.str();
}
string RecordNode::lid(){
  stringstream str;
/*  str<<"(" << val << ")"; */
  return str.str();
}
string RecordNode::lprint(){
  stringstream str;
  record_t::iterator record_it;
  for (record_it = __record->begin(); record_it != __record->end(); ++record_it) {
		if (CONSTVAL == record_it->nodetype) {
			str << record_it->get_val() << " ";
		} else {
			str << record_it->get_node()->lprint() << " ";
		}
  }

  return str.str();
}
ExpNode* RecordNode::clone(bool copyChildren) {
  return new RecordNode(*this);
};

void RecordNode::setValue(size_t field, EnvNode exp) {
#ifdef __DEBUG_1
  Assert(__record->size() > field, "Record should be big enough.");
#endif
  if (EXPNODE == exp.nodetype) {
	  merge_sym(exp.get_node());
	  exp.get_node()->inc_ref_count();
  }
  (*__record)[field] = exp;
}

/**
 * ListNode
 */
ListNode::ListNode() : ExpNode() {
  type = LIST;
  __sym = false;
}
ListNode::~ListNode() {

}
ListNode::ListNode(const ListNode& bn):
  ExpNode(bn), __rel(bn.__rel) {
  cout << "ListNode::ListNode(const ListNode& bn)" << endl;
}

void ListNode::accept(NodeVisitor& visitor)  { visitor.visit( *this ); }
string ListNode::get_namestr() const{
  stringstream str;
  str<< globalId <<"_C";
  return str.str();
}
string ListNode::lid(){
  stringstream str;
  // str<<"(" << val << ")";
  return str.str();
}
string ListNode::lprint() {
	stringstream str;
	relation_it rel_it;
	for (rel_it = __rel.begin(); rel_it != __rel.end(); ++rel_it) {
		if (*rel_it != NULL) {
			str << (*rel_it)->lprint() << ", ";
		} else {
			str << "NULL, ";
		}
	}
	// str<<id<<"= "<<"(" << val << ")";
	return str.str();
}
ExpNode* ListNode::clone(bool copyChildren) {
  return new ListNode(*this);
};

EnvNode ListNode::lookupRelationVal(int pkey, int field) {
#ifdef __DEBUG_1
  cout << "ListNode::lookupRelationVal: " << pkey << " at field " << field
       << endl;
#endif
  relation_it rel_it;
  for (rel_it = __rel.begin(); rel_it != __rel.end(); ++rel_it) {
#ifdef __DEBUG_1
    Assert(ExpNode::RECORD == (*rel_it)->type, "node must be record");
#endif
    RecordNode* record = (RecordNode*)*rel_it;
    EnvNode pkey_val = record->lookupField(0);
    if ((CONSTVAL == pkey_val.nodetype) && (pkey_val.get_val() == pkey)) {
      return record->lookupField(field);
    }
  }
  Assert(false, "Didn't find it.");
}

void ListNode::addExpNode(ExpNode* record) {
  merge_sym(record);
  __rel.push_back(record);

  /* Add this to the primary key index table. */
  if (ExpNode::RECORD == record->type) {
    EnvNode pkey = ((RecordNode*)record)->lookupField(0);
    if (CONSTVAL == pkey.nodetype) {
	    int pkey_val = pkey.get_val();
  	  if (__mapped_rel.find(pkey_val) == __mapped_rel.end()) {
	  	  __mapped_rel[pkey_val] = new relation_t();
	    }
  	  __mapped_rel[pkey_val]->push_back(record);
    }
  }
}

EnvNode ListNode::hasElem(ExpNode* en) {
  relation_it rel_it;
  ExpNode* has_elem = new ConstNode(false);

  for (rel_it = __rel.begin(); rel_it != __rel.end(); ++rel_it) {
    EnvNode eq_nodes = (*rel_it)->eqNode(en);
    if ((CONSTVAL == eq_nodes.nodetype) && (eq_nodes.get_val())) {
      return EnvNode(true);
    } else if (EXPNODE == eq_nodes.nodetype) {
      OrNode* on = new OrNode();
      on->set_parents(has_elem, eq_nodes.get_node());
      has_elem = on;
    }
  }
#ifdef __DEBUG_1
  cout << "ListNode::hasElem: Returning " << has_elem->lprint() << endl;
#endif
  return EnvNode(has_elem);
}

EnvNode ListNode::lookupSymVal(SymNode* en, size_t field) {
  relation_it rel_it;
  /* Ghetto hack. */
  ExpNode* eq_elem = new ConstNode(-1);

  for (rel_it = __rel.begin(); rel_it != __rel.end(); ++rel_it) {
    /* TODO: Dangureous. */
    RecordNode* rn = dynamic_cast<RecordNode*>(*rel_it);

    EnvNode fnode = rn->lookupField(0);
    ExpNode* field_node;
    
    EnvNode eq_nodes = EnvNode(-1);
    if (CONSTVAL == fnode.nodetype) {
      field_node = new ConstNode(fnode.get_val());
      eq_nodes = field_node->eqNode(en);
/*      if (0 == field_node->get_ref_count()) {
        delete(field_node);
      } */
    } else {
      field_node = fnode.get_node();
      eq_nodes = field_node->eqNode(en);
    }
    
    if ((CONSTVAL == eq_nodes.nodetype) && (eq_nodes.get_val())) {
      EnvNode cur = rn->lookupField(field);
      if (CONSTVAL == cur.nodetype) {
        eq_elem = new ConstNode(cur.get_val());
      } else {
        eq_elem = cur.get_node();
      }
    } else if (EXPNODE == eq_nodes.nodetype) {
      ExpNode* cur_eq = eq_nodes.get_node();

      ExpNode* field_node;
      EnvNode cur = rn->lookupField(field);
      if (CONSTVAL == cur.nodetype) {
        field_node = new ConstNode(cur.get_val());
      } else {
        field_node = cur.get_node();
      } 

      TernaryNode* on = new TernaryNode(cur_eq);
      on->set_parents(field_node, eq_elem);
      eq_elem = on;
    }
  }
  return EnvNode(eq_elem);
}
