/**
 *  This file defines different kinds of Boolean nodes.
 */
#include "ExpressionNodes.h" 

using namespace std;

InterNode::InterNode(const InterNode& bn): ExpNode(bn), nbits(bn.nbits) { }
InterNode::InterNode(){nbits = 1;}

int InterNode::get_nbits() const { return nbits; }
void InterNode::set_nbits(int n){ nbits = n; }

ExpNode::OutType InterNode::getOtype() const {
  if(otype != BOTTOM){
    return otype;
  }
  if(nbits>1){
    otype = INT;
  }else{
    otype = BOOL;
  }
  return otype;
}

string InterNode::get_namestr() const{
  stringstream str;
  if(globalId > -1)
    str<< globalId <<"__"<<get_tname();
  else{
    str<<"name_"<<abs(globalId)<<"_"<<"__"<<get_tname();

  }

  return str.str();
}

string InterNode::lprint(){
  stringstream str;
  str << "s" << globalId;
  return str.str();
}

string InterNode::lid() {
  return lprint();
}

/**
 * Symbolic nodes.
 * Their __sym flag should always be set to true.
 * The __name of symbolic nodes should correspond to the name of the symbolic
 * node in the front-end
 */
/* TODO: Figure out how variable names work with symbolic variables. */
SymNode::SymNode() : ConstNode() {
	type = SYM;
	__sym = true;
	__frontend_var = -1;
}

SymNode::SymNode(int frontend_var) : ConstNode() {
	type = SYM;
	__sym = true;
	__frontend_var = frontend_var;
}

SymNode::SymNode(const SymNode& bn): ConstNode(bn) { }

void SymNode::accept(NodeVisitor& visitor) {
  visitor.visit( *this );
}
ExpNode* SymNode::clone(bool copyChildren) {
  return new SymNode(*this);
};

void SymNode::addVars(set<int>& vars) {
  vars.insert(__frontend_var);
}

string SymNode::lprint() {
	stringstream str;
	if (type == ExpNode::SYM) {
			str << "s" << __frontend_var << "(" << this->getVal() << ")";
	} else {
			str << "(" << this->getVal() << ")";
	}
	return str.str();
}

yices_expr SymNode::to_yices(yices_context yctxt) {
   /* Update the Yices environment. */
  yices_var_decl v;
  // Make variable name
  stringstream str;
  str << "v" << __frontend_var;
  string cstr = str.str();

  /* Declare a new variable if necessary. */
  v = yices_get_var_decl_from_name(yctxt, (char*)cstr.c_str());
 
  if (v == 0) {
    // Get type
    yices_type intty = yices_mk_type(yctxt, "int");

    // Make variable declaration
    v = yices_mk_var_decl(yctxt, (char*)cstr.c_str(), intty);
  }

  return yices_mk_var_from_decl(yctxt, v);
}

string SymNode::get_yicesstr() const {
  stringstream str;
  str << "v" << __frontend_var;
  return str.str();
}

int SymNode::get_frontend_var() { return __frontend_var; }

EnvNode SymNode::eqNode(ExpNode* other) {
	if (other->getType() == ExpNode::SYM) {
		return
      EnvNode(globalId == other->globalId ||
				((__frontend_var >= 0) &&
				 __frontend_var == (dynamic_cast<SymNode*>(other))->get_frontend_var()));
	} else if (ExpNode::CONST == other->getType()) {
    EqNode* en = new EqNode();
    en->set_parents(this, other);
    return EnvNode(en);
  } else {
		return EnvNode(false);
	}
}
