#include "ExpressionNodes.h"
#include "ExpDAG.h"

#include <fstream>
#include <sstream>
#include <algorithm>

/**
 * ConstNode
 */
ConstNode::ConstNode() : ExpNode() {
  type = CONST; __val = -1;
}
ConstNode::ConstNode(int n) {
  type = CONST; __val = n;}
  ConstNode::ConstNode(const ConstNode& bn): ExpNode(bn), __val(bn.__val){
    //if(val == 13){ cout<<" surprise"<<endl; }
  }
void ConstNode::accept(NodeVisitor& visitor)  { visitor.visit( *this ); }
void ConstNode::setVal(int v){ __val = v; }
int ConstNode::getVal() const { return __val; }

string ConstNode::get_namestr() const{
  stringstream str;
  str<< globalId <<"_C";
  if( __val<0){
    str<<"m";
  }
  str<<abs(__val);
  return str.str();
}

string ConstNode::get_yicesstr() const{
  stringstream str;
  if (otype == ExpNode::BOOL) {
	  if (__val == 1) { str << "true"; }
	  else { str << "false"; }
  } else {
	  str << __val;
  }
  return str.str();
}

string ConstNode::lid(){
  stringstream str;
  str<<"(" << __val << ")";
  return str.str();
}
string ConstNode::lprint(){
  stringstream str;
  str<<"(" << __val << ")";
  return str.str();
}
ExpNode* ConstNode::clone(bool copyChildren) {
  return new ConstNode(*this);
};
ExpNode::OutType ConstNode::getOtype()const {
  if(otype != BOTTOM){
    return otype;
  }
  /* TODO: Umm??? */
  if(__val != 0 && __val != 1){
    otype = INT;
  }else{
    otype = BOOL;
  }
  return otype;
}
ExpNode::Type ConstNode::getType() const { return CONST; }

bool ConstNode::eqNode(ExpNode* other) {
	if (other->getType() == ExpNode::CONST) {
		return (__val == ((ConstNode*)other)->getVal());
	} else { return false; }
}

/**
 * RecordNode
 */
RecordNode::RecordNode() {
  type = RECORD;
  /* TODO */
  __record = new vector<ExpNode*>();
  __num_fields = 0;
  sym = false;
}
RecordNode::~RecordNode() {
  delete __record;
}
RecordNode::RecordNode(int numfields) {
#ifdef __DEBUG_1
	cout << "RecordNode::RecordNode: Creating record with " << numfields
		 << " fields." << endl;
#endif
	type = RECORD;
	__record = new vector<ExpNode*>(numfields, NULL);
	__num_fields = numfields;
	sym = false;
	cout << "done allocating" << endl;
}
RecordNode::RecordNode(const RecordNode& bn)
  : ExpNode(bn), __record(bn.__record), __num_fields(bn.__num_fields) { }

void RecordNode::accept(NodeVisitor& visitor)  { visitor.visit( *this ); }
string RecordNode::get_namestr() const{
  stringstream str;
  str<< globalId <<"_C";
/*  str<<abs(val); */
  return str.str();
}
string RecordNode::lid(){
  stringstream str;
/*  str<<"(" << val << ")"; */
  return str.str();
}
string RecordNode::lprint(){
  stringstream str;
  vector<ExpNode*>::iterator record_it;
  for (record_it = __record->begin(); record_it != __record->end(); ++record_it) {
	  str << (*record_it)->lprint() << " ";
  }

  return str.str();
}
ExpNode* RecordNode::clone(bool copyChildren) {
  return new RecordNode(*this);
};
ExpNode::OutType RecordNode::getOtype()const {
  if(otype != BOTTOM){
    return otype;
  } else {
    return RECORDTY;
  }
}

ExpNode::Type RecordNode::getType() const { return RECORD; }

record_t* RecordNode::get_record() { return __record; }

ExpNode* RecordNode::lookupField(int field) {
  Assert( field < __record->size()
        , "Must access a field in the record.");
  return (*__record)[field];
}

void RecordNode::setValue(int field, ExpNode* exp) {
#ifdef __DEBUG_1
  cout << "RecordNode:: setValue: " << field << " := " << exp->lprint() << endl;
#endif
  sym = sym || exp->sym;
  (*__record)[field] = exp;
}

/**
 * RelationNode
 */
RelationNode::RelationNode() {
  type = RELATION;
  __next_itid = 0;
  __num_fields = 0;
  sym = false;
}
RelationNode::~RelationNode() {

}
RelationNode::RelationNode(const RelationNode& bn):
  ExpNode(bn), __rel(bn.__rel), __next_itid(bn.__next_itid), __num_fields(bn.__num_fields) {
  cout << "RelationNode::RelationNode(const RelationNode& bn)" << endl;
}

void Tokenize(const string& str, vector<string>& tokens,
		const string& delimiters = " ") {
	// Skip delimiters at beginning.
	string::size_type lastPos = str.find_first_not_of(delimiters, 0);
	// Find first "non-delimiter".
	string::size_type pos = str.find_first_of(delimiters, lastPos);

	while (string::npos != pos || string::npos != lastPos) {
		// Found a token, add it to the vector.
		tokens.push_back(str.substr(lastPos, pos - lastPos));
		// Skip delimiters.  Note the "not_of"
		lastPos = str.find_first_not_of(delimiters, pos);
		// Find next "non-delimiter"
		pos = str.find_first_of(delimiters, lastPos);
	}
}

/* Construct a relation from file. */
/* TODO: Do something with numfields. */
RelationNode::RelationNode(int numfields, const char* file_name) {
#ifdef __DEBUG_1
	cout << "RelationNode::RelationNode: Reading from file " << file_name
			<< endl;
#endif
	__num_fields = numfields;
	__next_itid = 0;

	ifstream file;
	file.open(file_name);
	Assert(file.is_open(), "Was not able to open file.");

	int pkey = 0;
	while (!file.eof()) {
		/* Get current line. */
		string curline;
		getline(file, curline);
#ifdef __DEBUG_1
		cout << "RelationNode::RelationNode: Read line " << curline << endl;
#endif

		/* Scan current line. */
		vector<string> tokens;
		Tokenize(curline, tokens);
		if (tokens.size() == __num_fields) {
			RecordNode* curRecord = new RecordNode(__num_fields);
			__rel[pkey] = curRecord;

			vector<string>::iterator token_it;
			int field = 0;
			for (token_it = tokens.begin(); token_it != tokens.end(); ++token_it) {
				bool is_digit = true;

				string::iterator pch;
				for (pch = token_it->begin(); pch != token_it->end(); ++pch) {
					if (!isdigit(*pch)) {
						is_digit = false;
					}
				}

#ifdef __TEST
				Assert(curRecord != NULL, "Should have a primary key.");
#endif
				if (is_digit) {
					int cur_val = atoi(token_it->c_str());
					ConstNode* cnode = new ConstNode(cur_val);
					__rel[pkey]->setValue(field, cnode);
				} else {
					/* Otherwise insert symbolic value. */
#ifdef __DEBUG_1
					cout << "RelationNode::RelationNode: Adding symbolic "
						 << "value." << endl;
#endif
					SymNode* cnode = new SymNode();
					sym = true;
					__rel[pkey]->setValue(field, cnode);
				}
				++field;
			} // rof token_it
#ifdef __DEBUG_1
			cout << "RelationNode::RelationNode:: Read record "
				 << curRecord->get_name() << ": " << curRecord->lprint() << endl;
#endif
		}

		++pkey; /* Increment the primary key. */
	} // elihw reading file
	file.close();
}
RelationNode::RelationNode(int numfields) {
	  type = RELATION;
	  __next_itid = 0;
	  __num_fields = numfields;
	  sym = false;
}

void RelationNode::accept(NodeVisitor& visitor)  { visitor.visit( *this ); }
string RelationNode::get_namestr() const{
  stringstream str;
  str<< globalId <<"_C";
  /* if( val<0){
    str<<"m";
  } */
  // str<<abs(val);
  return str.str();
}
string RelationNode::lid(){
  stringstream str;
  // str<<"(" << val << ")";
  return str.str();
}
string RelationNode::lprint(){
  stringstream str;
  map<int, RecordNode*>::iterator rel_it;
  for (rel_it = __rel.begin(); rel_it != __rel.end(); ++rel_it) {
	  str << (rel_it->second)->lprint() << ", ";
  }
  // str<<id<<"= "<<"(" << val << ")";
  return str.str();
}
ExpNode* RelationNode::clone(bool copyChildren) {
  return new RelationNode(*this);
};
ExpNode::OutType RelationNode::getOtype()const {
  if(otype != BOTTOM){
    return otype;
  } else {
    return RELTY;
  }
}

ExpNode::Type RelationNode::getType() const { return RELATION; }

map<int, RecordNode*>& RelationNode::get_relation() {
	return __rel;
}

int RelationNode::getIterator() {
  int id = __next_itid;
  __iterators[__next_itid++] = __rel.begin();
  return id;
}

RecordNode* RelationNode::nextRecord(int itid) {
  relation_it cur_it = __iterators[itid];
#ifdef __DEBUG_1
  Assert(__iterators.find(itid) != __iterators.end()
        , "Iterator exists.")
#endif
  if (cur_it == __rel.end()) {
    return NULL;
  }
  RecordNode* cur_exp = (cur_it->second);
  __iterators[itid] = ++cur_it;
#ifdef __DEBUG_1
  cout << "RelationNode::nextRecord: next record is " << cur_exp->lprint()
	   << endl;
#endif
  return cur_exp;
}

ExpNode* RelationNode::lookupRelationVal(int pkey, int field) {
  if (__rel.find(pkey) != __rel.end()) {
    Assert(__rel[pkey] != NULL, "Record in relation cannot be null.");
    Assert(__rel[pkey]->getType() == RECORD
          , "Relation should contain records.");
    return __rel[pkey]->lookupField(field);
  } else {
    return NULL;
  }
}

void RelationNode::setRecord(int pkey, RecordNode* record) {
  Assert(record != NULL, "New record cannot be null.");
  sym = sym || record->sym;
  __rel[pkey] = record;
}

void RelationNode::setValue(int pkey, int field, ExpNode* exp) {
#ifdef __TEST
	Assert(__rel.find(pkey) != __rel.end(), "pkey should exist in relation");
	Assert(__rel[pkey] != NULL
			, "relation corresponding to primary key should not be null");
#endif
	sym == sym || exp->sym;
	__rel[pkey]->setValue(field, exp);
}

int RelationNode::get_num_fields() { return __num_fields; }
