/**
 * In this file we define the different parts of the backend environment.  We
 * have the following:
 * - Nodes, which can be:
 *      - Arithmetic operations.
 *      - Boolean operations.
 *      - Symbolic expressions.
 * - A mapping from expression ID's to
 * - 
 *
 * Questions:
 * - Should relations be different kinds of expressions than expressions?
 */
#ifndef __BACKEND_ENV_H
#define __BACKEND_ENV_H

#include <map>
#include <set>
#include "backend_defs.h"
#include "BackendTypes.h"
#include "ExpDAG.h"
#include "ExpDAGCreator.h"
#include "ExpressionNodes.h"

using namespace std;

class Relation;

typedef map<vname, ExpNode*> namemap_t;
typedef map<vname, relation_t*> relmap_t;

/**
 * Backend environment.
 */
class BackendEnv {
  private:
    /* The backend environment consists of the following:
     *  - __named_nodes: a mapping from front-end variable names to pointers to
     *      expression nodes.
     *  - __assertions: a mapping from expression ID's of the assertions.
     */
    namemap_t __named_nodes;

    vector<AssertNode*> __assertions;
    set<int> __constraint_vars;

    ExpDAG* __dag;
    ExpDAGCreator* __dag_creator;

  public:
    BackendEnv();

    ExpDAG* getDAG();
    ExpNode* new_node(ExpNode* mother, ExpNode* father, ExpNode::Type t);
    ExpNode* new_node(ExpNode* mother, ExpNode* father, ExpNode* node);
    ExpNode* get_node(eid_t eid);

    /* Named nodes. */
    /* TODO: Maybe get rid of this too? */
    /* TODO: I think the environment stores these variables to store a mapping
     * from original names to backend internal names.  Figure out if this is
     * true? */
    namemap_t& getNamedNodes();
    eid_t getEidByName(vname name);
    ExpNode* getExpByName(vname name);
    void setEidByName(vname name, eid_t eid); 
    void setExpByName(vname name, ExpNode* node);
    bool hasExpByName(vname name);
    void removeByName(vname name);
    void removeExpFromNames(ExpNode* exp);

    /**
     * Assertions.
     */
    void addConstraintVar(int newvar);
    set<int>& getVars();
    void addAssertion(AssertNode* new_assertion);
    void writeAssertionsToFile(string& file);

    /* Clears the environment. */
    void clear();
};

#endif /* __BACKEND_ENV_H */
