#include "DagCSE.h"
#include "SATSolver.h"
#include <sstream>




char tbl[64] = {'0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 
                'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 
				'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 
				'U', 'V',
				'W', 'X', 'Y', 'Z', 'a', 'b', 'c', 'd', 'e', 'f', 
                'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 
				'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z', 
				'@', '#'
};


int DagCSE::cheapStr(int id1, char op, int id2){
	int p = 0;
	char* tch = &tmpbuf[0];
	writeInt(tch, id1, p);
	tch[p] = op; p++;
	writeInt(tch, id2, p);
	tch[p] = 0;		
	return p;
	/*
	sprintf(tmpbuf, "%d%c%d",id1, op,id2);   	
 	ccode = tmpbuf;
	*/
}


void DagCSE::setStr(int id1, char op, int id2){
	int p = 0;
	char* tch = &tmpbuf[0];
	writeInt(tch, id1, p);
	tch[p] = op; p++;
	writeInt(tch, id2, p);
	tch[p] = 0;
	ccode = tch;		
	/*
	sprintf(tmpbuf, "%d%c%d",id1, op,id2);   	
 	ccode = tmpbuf;
	*/
}


DagCSE::DagCSE(ExpDAG& p_dag): dag(p_dag), tmpbuf(1000) Dtime(,stimer("string creation"), maptimer("map lookup"))
{
}

DagCSE::~DagCSE()
{
Dtime(stimer.print();)
Dtime(maptimer.print();)
	
}





void DagCSE::eliminateCSE(){
	int k=0;
	for(int i=0; i<dag.size(); ++i){
		// Get the code for this node. 		
		ExpNode* node = dag[i]; 

		ExpNode* cse = computeCSE(node);
		if(cse != node){
			Dout(cout<<"replacing "<<dag[i]->get_name()<<" -> "<<cse->get_name()<<endl );
			dag.replace(i, cse); 
		}
	}
	dag.removeNullNodes();
	dag.relabel();
	Dout(cout<<" end cse "<<endl);	
}

ExpNode* DagCSE::quickcse(int mid, int fid, ExpNode::Type t){		
	ExpNode* rv;
	switch(t){
		case ExpNode::AND:{
		int len = cheapStr(min(mid, fid), '&' ,max(mid, fid));
		if(cse_map.get(&tmpbuf[0], len, rv)){ return rv; }else{ return NULL; }
		}
		case ExpNode::OR:{
		int len = cheapStr(min(mid, fid), '|' ,max(mid, fid));
		if(cse_map.get(&tmpbuf[0], len, rv)){ return rv; }else{ return NULL; }
		}
		case ExpNode::NOT:
			{
				char* tch = &tmpbuf[0];
 				int p = 0;
				tch[p] = '!'; p++;
				writeInt(tch, mid, p);	
				tch[p]=0;
				if(cse_map.get(tch, p, rv)){ return rv; }else{ return NULL; }
			}

		default:
			return NULL;
	}
}

 void DagCSE::visit(  AndNode& node ){
 	Dtime(stimer.restart();)
 	
 	int mid = node.mother == NULL? -1: node.mother->globalId;
 	int fid = node.father == NULL? -1: node.father->globalId; 
 		
	setStr(min(mid, fid), '&' ,max(mid, fid));

 	Dtime(stimer.stop();)
 }
 void DagCSE::visit(  OrNode& node ){
 	Dtime(stimer.restart();)
 	int mid = node.mother == NULL? -1: node.mother->globalId;
 	int fid = node.father == NULL? -1: node.father->globalId;
	setStr(min(mid, fid), '|' ,max(mid, fid));

 	Dtime(stimer.stop();)
 }
 void DagCSE::visit(  SymNode& node ){
 	Dtime(stimer.restart();)
 	stringstream str;
 	str<<node.globalId;
 	ccode = str.str();
 	Dtime(stimer.stop();)
 }
 void DagCSE::visit(  NOT_node& node ){
 	Dtime(stimer.restart();)
 	
 	int mid = node.mother == NULL? -1: node.mother->globalId;
 	char* tch = &tmpbuf[0];
 	int p = 0;
	tch[p] = '!'; p++;
	writeInt(tch, mid, p);	
	tch[p]=0;
 	ccode = tch;
 	
 	Dtime(stimer.stop();)
 }
  /*
 void DagCSE::visit(  CTRL_node& node ){
 	Dtime(stimer.restart();)
 	stringstream str;
 	str<<node.globalId;
 	ccode = str.str();
 	Dtime(stimer.stop();)
 }
 */
 void DagCSE::visit(  PLUS_node& node ){
 	Dtime(stimer.restart();)
 	Assert( node.mother != NULL, "Null mother no longer allowed!!");
 	Assert( node.father != NULL, "Null father no longer allowed!!");

 	int mid = node.mother == NULL? -1: node.mother->globalId;
 	int fid = node.father == NULL? -1: node.father->globalId;
 	
 	setStr(min(mid, fid), '+' ,max(mid, fid));

 	
 	Dtime(stimer.stop();)
 }
 void DagCSE::visit(  TIMES_node& node ){
 	Dtime(stimer.restart();)
 	
 	int mid = node.mother == NULL? -1: node.mother->globalId;
 	int fid = node.father == NULL? -1: node.father->globalId; 	
	
	setStr(min(mid, fid), '*' ,max(mid, fid));

 	
 	Dtime(stimer.stop();)
 }
 
 void DagCSE::visit(  DIV_node& node ){
 	Dtime(stimer.restart();)
 	
 	int mid = node.mother == NULL? -1: node.mother->globalId;
 	int fid = node.father == NULL? -1: node.father->globalId;
 	
	setStr(min(mid, fid), '/' ,max(mid, fid));

 	
 	Dtime(stimer.stop();)
 }
 
 void DagCSE::visit(  NEG_node& node ){
 	Dtime(stimer.restart();)

 	int mid = node.mother == NULL? -1: node.mother->globalId; 	
 	char* tch = &tmpbuf[0];
	int p = 0;
	tch[p] = '-'; p++;
	writeInt(tch, mid, p);	
	tch[p]=0;
 	ccode = tch;
 	
 	Dtime(stimer.stop();)
 }
 
  void DagCSE::visit(  ConstNode& node ){
  	Dtime(stimer.restart();)
  	char* tch = &tmpbuf[0];
	int p = 0;
	tch[p] = '$'; p++;
	writeInt(tch, node.getVal(), p);	
	tch[p] = '$'; p++;
	tch[p]=0;  	
 	ccode = tch;
  	
 	Dtime(stimer.stop();)
 }
 
 
 void DagCSE::visit(  GT_node& node ){
 	Dtime(stimer.restart();)
 	
 	int mid = node.mother == NULL? -1: node.mother->globalId;
 	int fid = node.father == NULL? -1: node.father->globalId;
 	
 	
	setStr(fid, '<', mid);
 	
 	Dtime(stimer.stop();)
 }
 void DagCSE::visit(  GE_node& node ){
 	Dtime(stimer.restart();)
 	
 	int mid = node.mother == NULL? -1: node.mother->globalId;
 	int fid = node.father == NULL? -1: node.father->globalId;
 	
	setStr(fid, '{', mid);

 	Dtime(stimer.stop();)
 }
 
 void DagCSE::visit(  LT_node& node ){
 	Dtime(stimer.restart();)
 	
 	int mid = node.mother == NULL? -1: node.mother->globalId;
 	int fid = node.father == NULL? -1: node.father->globalId;
 	
 	setStr(mid, '<', fid);
 	
 	Dtime(stimer.stop();)
 }
 
 
 void DagCSE::visit(  LE_node& node ){
 	Dtime(stimer.restart();)
 	
 	int mid = node.mother == NULL? -1: node.mother->globalId;
 	int fid = node.father == NULL? -1: node.father->globalId;
 	
	setStr(mid, '{', fid);
 	
 	Dtime(stimer.stop();)
 }
 void DagCSE::visit(  EQ_node& node ){
 	Dtime(stimer.restart();)
 	
 	int mid = node.mother == NULL? -1: node.mother->globalId;
 	int fid = node.father == NULL? -1: node.father->globalId;
 	
 	
	setStr(min(mid, fid), '=', max(mid, fid));
 	 	
 	Dtime(stimer.stop();)
 }

/*
 void DagCSE::visit(  ACTRL_node& node ){
 	Dtime(stimer.restart();)
 	stringstream str;
 	for(int i=0; i<node.multi_mother.size(); ++i){
 		int mmid = node.multi_mother[i] == NULL? -1: node.multi_mother[i]->globalId;
 		str<<mmid<<",";
 	}
 	ccode = str.str();
 	Dtime(stimer.stop();)
 }
*/
 
 
 
 void DagCSE::visit(  AssertNode &node){
 	Dtime(stimer.restart();)
	int mid = node.mother == NULL? -1: node.mother->globalId;
	setStr(0, 'A', mid);
 	Dtime(stimer.stop();)
 }
 
 	
