/*
 * What is this file used for?
 */
#pragma once

#include "BackendTypes.h"
#include "ExpDAG.h"
#include "DagOptim.h"

class ExpDAGCreator
{
  /* Mapping from input variable names to expressions? */
	map<vname, ExpNode*> named_nodes;
	int new_names;
	int new_namesb;

	ExpDAG* const dag;
	DagOptim optim;
    void getMotherFather(const int mother,
					 const int father, /*OUT*/ExpNode*& mth, /*OUT*/ExpNode*& fth);

public:
	map<int, vector<ExpNode*> > methdparams;
	ExpDAGCreator(ExpDAG* p_dag);
public:
	virtual ~ExpDAGCreator(void);
  virtual ExpNode* create_const(int n);

  /* TODO: Do we still need inputs and outputs? */
  ExpNode* create_inputs(int n, const string& gen_name=string("INPUT"));

  /**
	Creates an N-bit output named gen_name connected to internal node nodeToOutput.
  */
  ExpNode* create_outputs(
      int n, ExpNode* nodeToOutput, const string& gen_name=string("OUTPUT"));
  ExpNode* create_outputs(int n, const string& gen_name=string("OUTPUT"));
  

  /** Produces a node called name.
  */
  ExpNode* get_node(const vname& name);  


  /* Associates the name source to the node  starg.
  */
  void alias(const vname& ssource,  ExpNode* starg);

  /**
	Create a new node from the string names of its parents.
  */
  ExpNode* new_node(ExpNode* mother, 
                      ExpNode* father, ExpNode::Type t);

  /* Used during initialization when creating ARITH nodes and ASSERT nodes.
   * thenode will always be an ARITH node or an ASSERT node.	The node thenode
   * should not be used after calls to this function, because this function may
   * optimize it away and deallocate it; so one must be very very careful to
   * always use the return value instead.  Also, all the fields of the arith
   * node should already be initialized. Otherwise, you'll get wrong behavior.
	*/
  ExpNode* new_node(ExpNode* mother, 
                      ExpNode* father, ExpNode* thenode);

  /*-------------------------------------*/

  ExpNode* optimizeAndAdd(ExpNode* node);
  
  bool has_alias(const vname& s){
    return named_nodes.find(s) != named_nodes.end();
  }

  ExpNode* get_alias(const vname& s){
    return get_node(s);
  }
  

  string new_name(){
	 int p = 0;
	char tmpbuf[200];
	tmpbuf[p] = 't'; p++;
	writeInt(tmpbuf, new_names, p);
	++new_names;
	tmpbuf[p] = 0;
    return tmpbuf;
  }
 string new_name(string& base){
 	if(base.length() < 2){
 		return new_name();	
 	}else{
	    stringstream str;
	    str<<base.substr(0,4)<<hex<<"_"<<new_names<<"__";
	    ++new_names;
	    return str.str();
 	}
  }


 void finalize(){
	 optim.cleanup(*dag);
 }

};
