#include "ExpressionNodes.h"
#include "ExpDAG.h"

#include <fstream>
#include <sstream>
#include <algorithm>

/**
 * ConstNode
 */
ConstNode::ConstNode() : ExpNode() {
  type = CONST; val = -1;
}
ConstNode::ConstNode(int n) {
  type = CONST; val = n;}
  ConstNode::ConstNode(const ConstNode& bn): ExpNode(bn), val(bn.val){
    //if(val == 13){ cout<<" surprise"<<endl; }
  }
void ConstNode::accept(NodeVisitor& visitor)  { visitor.visit( *this ); }
void ConstNode::setVal(int v){ val = v; }
int ConstNode::getVal() const { return val; }

string ConstNode::get_namestr() const{
  stringstream str;
  str<<__name<<"_C";
  if( val<0){
    str<<"m";
  }
  str<<abs(val);
  return str.str();
}

string ConstNode::get_yicesstr() const{
  stringstream str;
  str << val;
  return str.str();
}

string ConstNode::lid(){
  stringstream str;
  str<<"(" << val << ")";
  return str.str();
}
string ConstNode::lprint(){
  stringstream str;
  str<<id<<"= "<<"(" << val << ")";
  return str.str();
}
ExpNode* ConstNode::clone(bool copyChildren) {
  return new ConstNode(*this);
};
ExpNode::OutType ConstNode::getOtype()const {
  if(otype != BOTTOM){
    return otype;
  }
  /* TODO: Umm??? */
  if(val != 0 && val != 1){
    otype = INT;
  }else{
    otype = BOOL;
  }
  return otype;
}
ExpNode::Type ConstNode::getType() const { return CONST; }

/**
 * BoolNode
 */
BoolNode::BoolNode() : ExpNode() {
  type = BOOLCONST; __val = false;
}
BoolNode::BoolNode(bool n) {
  type = BOOLCONST; __val = n;
}
BoolNode::BoolNode(const BoolNode& bn): ExpNode(bn), __val(bn.__val){
  //if(val == 13){ cout<<" surprise"<<endl; }
}
void BoolNode::accept(NodeVisitor& visitor)  { visitor.visit( *this ); }
void BoolNode::setVal(bool v){ __val = v; }
int BoolNode::getVal() const { return __val; }

string BoolNode::get_namestr() const{
  stringstream str;
  str<<__name<<"_C";
  return str.str();
}

string BoolNode::get_yicesstr() const{
  stringstream str;
  if (__val) {
    str << "true";
  } else {
    str << "false"; 
  }
  return str.str();
}

string BoolNode::lid(){
  stringstream str;
//  str<<"(" << val << ")";
  return str.str();
}
string BoolNode::lprint(){
  stringstream str;
//  str<<id<<"= "<<"(" << val << ")";
  return str.str();
}
ExpNode* BoolNode::clone(bool copyChildren) {
  return new BoolNode(*this);
};
ExpNode::OutType BoolNode::getOtype()const {
  if(otype != BOTTOM){
    return otype;
  }
  return BOOL;
}

ExpNode::Type BoolNode::getType() const { return BOOLCONST; }

/**
 * RecordNode
 */
RecordNode::RecordNode() {
  type = RECORD;
  /* TODO */
  __record = new vector<eid_t>(100);
}
RecordNode::~RecordNode() {
  delete __record;
}
RecordNode::RecordNode(const RecordNode& bn)
  : ExpNode(bn), __record(bn.__record) { }
void RecordNode::accept(NodeVisitor& visitor)  { visitor.visit( *this ); }
string RecordNode::get_namestr() const{
  stringstream str;
  str<<__name<<"_C";
/*  str<<abs(val); */
  return str.str();
}
string RecordNode::lid(){
  stringstream str;
/*  str<<"(" << val << ")"; */
  return str.str();
}
string RecordNode::lprint(){
  stringstream str;
/*   str<<id<<"= "<<"(" << val << ")"; */
  return str.str();
}
ExpNode* RecordNode::clone(bool copyChildren) {
  return new RecordNode(*this);
};
ExpNode::OutType RecordNode::getOtype()const {
  if(otype != BOTTOM){
    return otype;
  } else {
    return RECORDTY;
  }
}

ExpNode::Type RecordNode::getType() const { return RECORD; }

eid_t RecordNode::lookupField(int field) {
  Assert( field < __record->size()
        , "Must access a field in the record.");
  return (*__record)[field];
}

void RecordNode::setValue(int field, eid_t eid) {
#ifdef __DEBUG_1
  cout << "RecordNode:: setValue: " << field << " := " << eid << endl;
#endif
  (*__record)[field] = eid;
#ifdef __TEST
  printf("RecordNode::setValue: Begin test...\n");
  Assert( __record->size() > field
        , "Record size should be at least as big as the number of fields.");
  int lookup_eid = this->lookupField(field);
  stringstream ss;
  ss << eid << " != " << lookup_eid;
  string neq_str = ss.str();
  Assert( lookup_eid == eid, neq_str);
  printf("RecordNode::setValue: End test.\n");
#endif
}

void RecordNode::addValue(eid_t eid) {
  __record->push_back(eid);
}

/**
 * RelationNode
 */
RelationNode::RelationNode() {
#ifdef __DEBUG_1
  cout << "RelationNode::RelationNode: Constructor." << endl;
#endif
  type = RELATION;
  __next_itid = 0;
}
RelationNode::~RelationNode() {

}
RelationNode::RelationNode(const RelationNode& bn):
  ExpNode(bn), __rel(bn.__rel), __next_itid(bn.__next_itid) {
#ifdef __DEBUG_1
  cout << "RelationNode::RelationNode(const RelationNode& bn)" << endl;  
#endif
}

/* Construct a relation from file. */  
RelationNode::RelationNode(const string& file_name) {
  Assert(false, "unimplemented: read from file");
#ifdef __DEBUG_1
  cout << "RelationNode::RelationNode: Reading from file " << file_name << endl;
#endif

  ifstream file;
  file.open("file.txt");
  Assert(file.is_open(), "Was not able to open file.");

  while(!file.eof()) {
    /* Get current line. */
    string curline;
    getline(file, curline);
    
    bool isPkey = true;
    RecordNode* curRecord = NULL;
    int pkey = 0;

    /* Scan current line. */
    char* curl = new char[curline.size()];
    for (int i = 0; i < curline.size(); ++i) {
      curl[i] = curline[i];
    }
    char* pch = strtok(curl, " \t");
    while(pch != NULL) {
      bool is_digit = true;
      for (; *pch != '\0'; ++pch) {
        if (!isdigit(*pch)) { is_digit = false; }
      }

      if (isPkey) {
        Assert(is_digit, "Primary key should be digit.");
        /* TODO: Relax this requirement. */
        int key = atoi(pch);
        __rel[key] = new RecordNode();
        pkey = key;
        isPkey = false;
      } else {
#ifdef __TEST
          Assert(curRecord != NULL, "Should have a primary key.");
#endif
        if (is_digit) {
          int cur_val = atoi(pch);
          ConstNode* cnode = new ConstNode(cur_val);
          __rel[pkey]->addValue(cnode->get_name());
        } else {
          /* Otherwise insert symbolic value. */
          SymNode* cnode = new SymNode();
          __rel[pkey]->addValue(cnode->get_name());
        }
      }
      pch = strtok(NULL, " \t");
    }
  }
  file.close();
}

void RelationNode::accept(NodeVisitor& visitor)  { visitor.visit( *this ); }
string RelationNode::get_namestr() const{
  stringstream str;
  str<<__name<<"_C";
  /* if( val<0){
    str<<"m";
  } */
  // str<<abs(val);
  return str.str();
}
string RelationNode::lid(){
  stringstream str;
  // str<<"(" << val << ")";
  return str.str();
}
string RelationNode::lprint(){
  stringstream str;
  // str<<id<<"= "<<"(" << val << ")";
  return str.str();
}
ExpNode* RelationNode::clone(bool copyChildren) {
  return new RelationNode(*this);
};
ExpNode::OutType RelationNode::getOtype()const {
  if(otype != BOTTOM){
    return otype;
  } else {
    return RELTY;
  }
}

ExpNode::Type RelationNode::getType() const { return RELATION; }

int RelationNode::getIterator() {
  int id = __next_itid;
  __iterators[__next_itid++] = __rel.begin();
#ifdef __DEBUG_1
  cout << "RelationNode::getIterator: Got iterator with ID " << id
       << ", should have " << __rel.size() << " elements." << endl;
#endif
  return id;
}

eid_t RelationNode::nextRecord(int itid) {
  relation_it cur_it = __iterators[itid];
#ifdef __DEBUG_1
  cout << "RelationNode::nextRecord: next record for " << itid << endl;
  Assert(__iterators.find(itid) != __iterators.end()
        , "Iterator exists.")
#endif
  if (cur_it == __rel.end()) {
    return -1;
  }
  eid_t cur_eid = (cur_it->second)->get_name();
  __iterators[itid] = ++cur_it;
#ifdef __DEBUG_1
  cout << "RelationNode::nextRecord: next record has ID " << cur_eid << endl;
#endif
  return cur_eid;
}

eid_t RelationNode::lookupRelationVal(int pkey, int field) {
  if (__rel.find(pkey) != __rel.end()) {
    Assert(__rel[pkey] != NULL, "Record in relation cannot be null.");
    Assert(__rel[pkey]->getType() == RECORD
          , "Relation should contain records.");
    printf("I get here.\n");
    return __rel[pkey]->lookupField(field);
  } else {
    return -1;
  }
}

void RelationNode::setRecord(int pkey, RecordNode* record) {
  Assert(record != NULL, "New record cannot be null.");
  __rel[pkey] = record;
}

void RelationNode::setValue(int pkey, int field, eid_t eid) {
#ifdef __DEBUG_1
  cout << "RelationNode::setValue: " << pkey << ", " << field << " := " << eid
       << endl;
#endif
  /*
  Assert (__rel.find(pkey) == __rel.end(), "value should not already exist.");
#ifdef __DEBUG_1
  cout << "RelationNode::setValue: Creating new record at primary key " << pkey
       << endl;
#endif
  __rel[pkey] = new RecordNode();
*/
#ifdef __TEST
  Assert(__rel.find(pkey) != __rel.end(), "pkey should exist in relation");
  Assert(__rel[pkey] != NULL
        , "relation corresponding to primary key should not be null");
#endif
 
  __rel[pkey]->setValue(field, eid);

#ifdef __DEBUG_1
  cout << "RelationNode::setValue: Finished setting value for " << pkey << endl;
#endif
}
