#include "NodeEvaluator.h"

NodeEvaluator::NodeEvaluator(map<string, ExpDAG*>& functionMap_p, ExpDAG& bdag_p):
functionMap(functionMap_p), trackChange(false), failedAssert(false), bdag(bdag_p)
{
	values.resize(bdag.size());
	changes.resize(bdag.size(), false);

}


NodeEvaluator::~NodeEvaluator(void)
{
}



void NodeEvaluator::visit( AndNode& node ){
	setbn(node, b(*node.mother) && b(*node.father));
}
void NodeEvaluator::visit( OrNode& node ){
	setbn(node, b(*node.mother) || b(*node.father));
}
void NodeEvaluator::visit( NOT_node& node ){
	setbn(node, !b(*node.mother));
}
void NodeEvaluator::visit( PLUS_node& node ){
	setbn(node, i(*node.mother) + i(*node.father));
}
void NodeEvaluator::visit( TIMES_node& node ){
	setbn(node, i(*node.mother) * i(*node.father));
}
void NodeEvaluator::visit( DIV_node& node ){
	int tt = i(*node.father);
	setbn(node, tt == 0 ? 0 : (i(*node.mother) / tt));
}
void NodeEvaluator::visit( NEG_node& node ){
	setbn(node, -i(*node.mother));
}
void NodeEvaluator::visit( ConstNode& node ){
	setbn(node, node.getVal());
}
void NodeEvaluator::visit( GT_node& node ){
	setbn(node, i(*node.mother) > i(*node.father));
}
void NodeEvaluator::visit( GE_node& node ){
	setbn(node, i(*node.mother) >= i(*node.father));
}
void NodeEvaluator::visit( LT_node& node ){
	setbn(node, i(*node.mother) < i(*node.father));
}
void NodeEvaluator::visit( LE_node& node ){
	setbn(node, i(*node.mother) <= i(*node.father));
}
void NodeEvaluator::visit( EQ_node& node ){
	setbn(node, i(*node.mother) == i(*node.father));
}

void NodeEvaluator::visit( SymNode& node ){
    setbn(node, i(*node.mother) == i(*node.father));
}
/*!
    multi-mother[0] = old-value;
    multi-mother[1] = new-value;
    if( mother == quant ) return multi-mother[1]; else return multi-mother[0];		
*/
void NodeEvaluator::visit( AssertNode &node){
	bool t = b(*node.mother);
	failedAssert = failedAssert || !t;
	setbn(node, t );
}	


bool NodeEvaluator::run(VarStore& inputs_p){
	inputs = &inputs_p;
	int i=0;
	failedAssert = false;
	for(ExpDAG::iterator node_it = bdag.begin(); node_it != bdag.end(); ++node_it, ++i){				
		(*node_it)->accept(*this);
	}
	return failedAssert;
}

void NodeEvaluator::display(ostream& out){
	for(int i=0; i<values.size(); ++i){
		cout<<"AVALS= "<<bdag[i]->lprint()<<"	v="<<values[i]<<endl;
	}
}


int NodeEvaluator::scoreNodes(){
	int i=0;
	int maxcount = -10;
	int highest;
	int nconsts = 0;
	for(vector<bool>::iterator it = changes.begin(); it != changes.end(); ++it, ++i){
		ExpNode* ni = bdag[i];
		if(!*it && ni->type != ExpNode::CONST){
			++nconsts;
			int count = 0;
			if(count > maxcount + 3){
				highest = i;
				maxcount = count;
			}
		}
	}
	cout<<"nconstants  = "<<nconsts<<"max count = "<<maxcount<<" highest = "<<highest<<endl;
	cout<<bdag[highest]->lprint()<<endl;
	/*
	for(child_iter cit = bdag[highest]->children.begin(); cit != bdag[highest]->children.end(); ++cit){
		cout<<(*cit)->lprint()<<endl;
	}*/
	return highest;
}
