#include <fstream>
#include "BackendEnv.h"

using namespace std;

/**
 * Construct a back-end environment.
 */
BackendEnv::BackendEnv() {
#ifdef __DEBUG_1
  cout << "BackendEnv::BackendEnv(): Constructor called." << endl;
#endif
  __dag = new ExpDAG();
  __dag_creator = new ExpDAGCreator(__dag);
}

/**
 * DAG functions.
 */
ExpDAG* BackendEnv::getDAG() { return __dag; }
ExpNode* BackendEnv::new_node(
    ExpNode* mother, ExpNode* father, ExpNode::Type t) {
	ExpNode* new_node = __dag_creator->new_node(mother, father, t);
	return new_node;
}
ExpNode* BackendEnv::new_node(
    ExpNode* mother, ExpNode* father, ExpNode* newnode) {
	ExpNode* new_node = __dag_creator->new_node(mother, father, newnode);
	return new_node;
}
ExpNode* BackendEnv::get_node(eid_t eid) {
  return __dag_creator->get_node(eid);
}

/**
 * Named nodes mapping variable names to nodes.
 */
namemap_t& BackendEnv::getNamedNodes() {
  return __named_nodes;
}

eid_t BackendEnv::getEidByName(vname name) {
  return __named_nodes[name]->get_name();
}

ExpNode* BackendEnv::getExpByName(vname name) {
  return __named_nodes[name];
}

void BackendEnv::setEidByName(vname name, eid_t eid) {
  ExpNode* node = __dag_creator->get_node(eid);
#ifdef __TEST
  Assert(node != NULL, "Should have node with given EID.");
  Assert(node->get_name() == eid, "Node we got should have right name.");
#endif
  __named_nodes[name] = node;
}

void BackendEnv::setExpByName(vname name, ExpNode* node) {
  __named_nodes[name] = node;
}

bool BackendEnv::hasExpByName(vname name) {
  return (__named_nodes.find(name) != __named_nodes.end());
}

void BackendEnv::removeByName(vname name) {
  __named_nodes.erase(name);
}

void BackendEnv::removeExpFromNames(ExpNode* enode) {
  namemap_t::iterator it = __named_nodes.find(enode->get_name());
  if(it != __named_nodes.end() && it->second==enode){
    __named_nodes.erase(it);
  }
}

void BackendEnv::addRecordNodes(RelationNode* relnode) {
	map<int, RecordNode*>::iterator rel_it = (relnode->get_relation()).begin();
	for (; rel_it != (relnode->get_relation()).end(); ++ rel_it) {
		__dag_creator->alias((rel_it->second)->get_name(), rel_it->second);
	}
}


/**
 * Node constructors.
 */

/**
 *  Assertions.
 */
set<int>& BackendEnv::getVars() {
#ifdef __DEBUG_1
  cout << "BackendEnv::getVars: Returning " << __constraint_vars.size()
	   << " variables." << endl;
#endif
  return __constraint_vars;
}

void BackendEnv::addConstraintVar(int var) {
  __constraint_vars.insert(var);
}

void BackendEnv::addAssertion(AssertNode* new_assertion) {
  __assertions.push_back(new_assertion);
}

void BackendEnv::writeAssertionsToFile(string& filename) {
#ifdef __DEBUG_1
  cout << "BackendEnv::writeAssertionsToFile: Writing to " << filename << endl;
#endif
  /* Open file for writing. */
  ofstream file;
  file.open(filename.c_str());

  /* Write variable declarations to file. */
  set<int>::iterator var_it = __constraint_vars.begin();
  for (; var_it != __constraint_vars.end(); ++var_it) {
    file << "(define v" << *var_it << ":: real)\n";
  }

  /* Write assertions themselves to file. */
  vector<AssertNode*>::iterator assert_it = __assertions.begin();
  for (; assert_it != __assertions.end(); ++assert_it) {
    file << (*assert_it)->get_yicesstr() << endl;
  }
  file << "(check)" << endl;

  file.close();
}

/**
 * Other.
 */
void BackendEnv::printEnv() {
	cout << "BackendEnv::printEnv: Printing environment..." << endl;
	map<vname, ExpNode*>::iterator env_it;
	for (env_it = __named_nodes.begin(); env_it != __named_nodes.end(); ++env_it) {
		cout << env_it->first << " |-> " << ((env_it->second)->lprint()) << endl;
	}
}

void BackendEnv::clear() {
  __named_nodes.clear();
  __assertions.clear();
  __constraint_vars.clear();
  __dag->clear();
}
