#include "ExpressionNodes.h"

/**
 * Assertions.
 */
AssertNode::AssertNode () : isHardAssert(false) {
  type = ASSERT;
}

AssertNode::AssertNode(ExpNode* aexp) : isHardAssert(false) {
  type = ASSERT;
  __assert_exp = aexp;
}

AssertNode::AssertNode(const AssertNode& bn): ExpNode(bn) {
  isHardAssert = bn.isHardAssert;  msg = bn.msg;
}

void AssertNode::accept (NodeVisitor &visitor)  { visitor.visit (*this); }

ExpNode* AssertNode::clone(bool copyChildren) {
  return new AssertNode(*this);
};

void AssertNode::makeHardAssert(){ isHardAssert = true; }
bool AssertNode::isHard() const { return isHardAssert ; }
void AssertNode::setMsg(const string& pmsg){ msg = pmsg; }
const string& AssertNode::getMsg()const{ return msg; }
string AssertNode::lprint(){
  stringstream str;
  str<<id<<"= ASSERT";

  str<<mother->lid()<<" : "<<msg;		
  return str.str();
}

ExpNode::Type AssertNode::getType() const { return ASSERT; }

string AssertNode::get_yicesstr() const {
  stringstream str;
#ifdef __TEST
  Assert( __assert_exp->getType() != ExpNode::ASSERT
		, "Cannot assert and assert node.");
#endif
  str << "(assert " << __assert_exp->get_yicesstr() << ")";
  return str.str();
}

void AssertNode::addVars(set<int>& varset) {
	__assert_exp->addVars(varset);
}

/**
 * Predicated nodes.
 */
TernaryNode::TernaryNode() {
  type = TERNARY;
  __guard = NULL;
}

TernaryNode::TernaryNode(ExpNode* guard) {
  type = TERNARY;
  __guard = guard;
}

TernaryNode::TernaryNode(const TernaryNode& bn) : ExpNode(bn) {
  __guard = bn.__guard;
}

void TernaryNode::accept (NodeVisitor &visitor) { visitor.visit(*this); }

ExpNode* TernaryNode::clone(bool copyChildren) {
  return new TernaryNode(*this);
}

string TernaryNode::lprint(){
  stringstream str;
  str<<id<<"= TERNARY";

  str<<mother->lid();
  return str.str();
} 

ExpNode::Type TernaryNode::getType() const { return TERNARY; }

string TernaryNode::get_yicesstr() const {
  stringstream str;
  str << "(if (" << __guard->get_yicesstr() << ")";
  str << mother->get_yicesstr() << " " << father->get_yicesstr() << ")";
  return str.str();
}

void TernaryNode::addVars(set<int>& varset) {
	__guard->addVars(varset);
	mother->addVars(varset);
	father->addVars(varset);
}
