#include "ExpressionNodes.h"
#include "ExpDAG.h"

#include <sstream>
#include <algorithm>

/**
 * ExpNode base class definitions.
 */
int ExpNode::__s_next_global_id = 0;

#ifdef SCHECKMEM
set<ExpNode*> ExpNode::allocated;
#endif 


ExpNode::ExpNode():globalId(__s_next_global_id++), mother(NULL), layer(0), father(NULL), flag(0), id(-1), ion_pos(0), otype(BOTTOM)
{ 
#ifdef SCHECKMEM
  allocated.insert(this);
#endif
  __name = __s_next_global_id; /* TODO: Is this what we want? */
  // cout << "next global ID: " << __name << endl;
}  
ExpNode::ExpNode(const ExpNode& bn):globalId(__s_next_global_id++), mother(bn.mother), layer(bn.layer), 
  __name(bn.__name), father(bn.father), 
  flag(bn.flag), id(bn.id), ion_pos(bn.ion_pos), 
  otype(bn.otype), type(bn.type)
{ 
#ifdef SCHECKMEM
  allocated.insert(this);
#endif
}

ExpNode::~ExpNode(){
#ifdef SCHECKMEM
  allocated.erase(this);
#endif
}

void ExpNode::resetId() { globalId = __s_next_global_id++; }

ExpNode::OutType ExpNode::joinOtype(OutType t1, OutType t2) const{
  if(t1 == BOTTOM){ return t2; }
  if(t2 == BOTTOM){ return t1; }
  if( t2 == t1 ){
    return t1;
  }else{
    return INT;
  }
}

void ExpNode::neighbor_replace(ExpNode* replacement){
  // ExpNode* onode = this;
	Assert(false, "ExpNode::neighbor_replace not implemented.");
}

void ExpNode::switchInputs(ExpDAG& bdag, map<ExpNode*, ExpNode*>& replacements){
  if(father != NULL){
    if(replacements.count(father) > 0){
      father = replacements[father];
    }		
  }
  if(mother != NULL){
    if(replacements.count(mother) > 0){
      mother = replacements[mother];
    }
  }
}

void ExpNode::redirectParentPointers(ExpDAG& oribdag, const vector<const ExpNode*>& bdag, bool setChildrn, ExpNode* childToInsert){
  if(father != NULL){
    Assert( father->id != -22, "This node should not exist anymore");
    if( oribdag.checkNodePosition(father) ){
      //If the father was from the original bdag, then we switch to the father in the new bdag.
      father = (ExpNode*) bdag[father->id];
    }
  }
  if(mother != NULL){
    Assert( mother->id != -22, "This node should not exist anymore");
    if( oribdag.checkNodePosition(mother)){
      mother = (ExpNode*) bdag[mother->id];
    }
  }
}

void ExpNode::redirectPointers(ExpDAG& oribdag, const vector<const ExpNode*>& bdag){
  redirectParentPointers(oribdag, bdag, false, NULL);

  /*
     for(child_iter child = tchild.begin(); child != tchild.end(); ++child){
     Assert( (*child)->id != -22, "This node should not exist anymore");
     if( oribdag.checkNodePosition((*child)) ){
     children.insert((ExpNode*) bdag[ (*child)->id ] );
     }else{
     children.insert(*child);
     }
     }
     */
}

string ExpNode::get_tname() const{
  switch(type){
    case PLUS: return "PLUS";
    case TIMES: return "TIMES";
    case DIV: return "DIV";
    case NEG: return "NEG";
    case CONST: return "CONST";
    case BOOLCONST: return "BOOLCONST";
    case SYM : return "SYM";
    case RELATION: return "RELATION";
    case RECORD: return "RECORD";
    case TERNARY: return "TERNARY";
    case AND: return "AND";
    case OR: return "OR";
    case NOT: return "NOT";
    case GT: return "GT";
    case GE: return "GE";
    case LT: return "LT";
    case LE: return "LE";
    case EQ: return "EQ";
    case ASSERT: return "ASSERT";
  }
  cout<<"ABOUT TO ABORT BECAUSE OF "<<__name<<"  "<<type<<endl;
  throw BasicError("Err", "Err");
}

string ExpNode::get_sym() const{
  switch(type){
    case PLUS: return "+";
    case TIMES: return "*";
    case DIV: return "/";
    case NEG: return "-";
    case CONST: return "C";
    case BOOLCONST: return "B";
    case SYM: return "S";
    case RELATION: return "R";
    case RECORD: return "r";
    case TERNARY: return "T";
    case AND: return "&";
    case OR: return "|";
    case NOT: return "!";
    case GT: return ">";
    case GE: return ">=";
    case LT: return "<";
    case LE: return "<=";
    case EQ: return "==";
    case ASSERT: return "ASSERT";
  }
  cout<<"ABOUT TO ABORT BECAUSE OF "<<__name<<"  "<<type<<endl;
  throw BasicError("Err", "Err");
}

void ExpNode::printval() {
	cout << "Printing expression " << this->get_name() << endl;
	cout << this->lprint() << endl;
}

string ExpNode::lprint(){
  stringstream str;
  str<<id<<"= ";
  if(mother != NULL){
    str<<mother->lid()<<" ";
  }
  str << get_sym()<<" ";
  if(father != NULL){
    str<<father->lid()<<" ";
  }
  return str.str();
}

string ExpNode::lid(){
  stringstream str;
  str<<id;
  return str.str();
}

vname ExpNode::get_name() { return __name; }
void ExpNode::set_name(vname name) { __name = name; }

/**
 * Serialization.
 */
string ExpNode::get_namestr() const {
  stringstream str;
  if(__name > -1)
    str<<__name<<"__"<<get_tname();
  else{      
    str<<"name_"<<abs(id)<<"__"<<get_tname();

  }    
  //str<<"_"<<this;
  //str<<":"<<id;
  Assert( id != -22, "This is a corpse. It's living gargabe "<<str.str()<<" id ="<<id );
  return str.str();
}

/**
 * The parent Yices-serialization function should never actually get called.
 */
string ExpNode::get_yicesstr() const {
	cout << "ExpNode::get_yicesstr: Type " << this->getType() << endl;
  Assert(false, "Yices serialization should be defined for this node.");
}

ExpNode::OutType ExpNode::getOtype() const{
  Assert(false, "This shouldn't get called");
  return BOOL;
}

ExpNode::Type ExpNode::getType() const{
  Assert(false, "Somebody didn't implement getType.");
  return AND;
}

void ExpNode::set_layer(bool isRecursive){
  if(!isRecursive){ layer = 0;}
  if(mother != NULL){
    if(isRecursive && mother->layer < 0){ mother->set_layer(isRecursive); }
    layer = mother->layer + 1;
  }
  if(father != NULL){
    if(isRecursive && father->layer < 0){
      father->set_layer(isRecursive); 
    }
    if(father->layer >= layer){
      layer = father->layer + 1;
    }
  }
}

int ExpNode::do_dfs(int idx){
  Assert( id != -22, "This node should be dead (dfs) "<<this->get_namestr());
  if( flag != 0)
    return idx;
  flag = 1;    
  /* TODO: Change this to do DFS over parents. */
  Assert(idx >= 0, "This shouldn't happen");
  id = idx;
  return idx-1;
}

int ExpNode::back_dfs(int idx){
  Assert( id != -22, "This node should be dead (back_dfs)"<<this->get_namestr());
  if( flag != 0)
    return idx;
  flag = 1;    
  if(father != NULL){
    idx = father->back_dfs(idx);
  }
  if(mother != NULL){
    idx = mother->back_dfs(idx);
  }
  id = idx;
  return idx+1;
}

void ExpNode::replace_parent(const ExpNode * oldpar, ExpNode* newpar){
  Assert( oldpar != NULL, "oldpar can't be null");
  Assert( newpar != NULL, "oldpar can't be null");
  if(father == oldpar){
    father = newpar;
    //oldpar->remove_child(this);
    //we assume the old parent is going to be distroyed, so we don't
    //bother modifying it.
  }
  if(mother == oldpar){
    mother = newpar;
    //oldpar->remove_child(this);
    //we assume the old parent is going to be distroyed, so we don't
    //bother modifying it.
  }
}

void ExpNode::printSubDAG(ostream& out, set<const ExpNode* >& s)const{
  s.insert(this);
  if( father != NULL && s.count(father) == 0 ){
    father->printSubDAG(out, s);
  }	
  if(mother != NULL && s.count(mother) == 0){
    mother->printSubDAG(out, s);
  }
  outDagEntry(out);	
}

void ExpNode::outDagEntry(ostream& out) const{
  if( father != NULL){
    out<<" "<<father->get_namestr()<<" -> "<<get_namestr()<<" ; "<<endl;
  }
  if( mother != NULL){
    out<<" "<<mother->get_namestr()<<" -> "<<get_namestr()<<" ; "<<endl;
  }
  if(father == NULL && mother == NULL){
    out<<"// orphan node: "<<get_namestr()<<" ; "<<endl ;
  }
}



void ExpNode::addVars(set<int>& vars) {
	/*
	if (mother != NULL) {
		mother->addVars(vars);
	}
	if (father != NULL) {
		father->addVars(vars);
	}
	*/
  return;
}

/**
 * Equality testing.
 */
bool ExpNode::eqNodes(ExpNode* n1, ExpNode* n2) {
  if (n1->getType() == n2->getType()) {
    if (n1->getType() == CONST) {
      return ((ConstNode*)n1)->getVal() == ((ConstNode*)n2)->getVal();
    } else if (n1->getType() == SYM) {
      return n1->get_name() == n2->get_name();
    } else {
      /* TODO */
      return false;
    }
  } else {
    return false;
  }
}

/**
 * Binary nodes.
 */
/**
 * AndNode
 */
AndNode::AndNode(const AndNode& b): ExpNode() { }
AndNode::AndNode(){ type = AND; }
void AndNode::accept(NodeVisitor& visitor) {
  visitor.visit( *this );
}
ExpNode* AndNode::clone(bool copyChildren = true) {
  return new AndNode(*this);
};
ExpNode::OutType AndNode::getOtype() const { return BOOL; }
ExpNode::Type AndNode::getType() const { return AND; }

string AndNode::get_yicesstr() const {
	stringstream s;
	s << "(" << mother->get_yicesstr() << " && " << father->get_yicesstr()
			<< ")";
	return s.str();
}

void AndNode::addVars(set<int>& varset) {
	mother->addVars(varset);
	father->addVars(varset);
}


/**
 * OrNode
 */
OrNode::OrNode(){ type = OR; }
OrNode::OrNode(const OrNode& bn): ExpNode(bn){ }
void OrNode::accept(NodeVisitor& visitor)  { visitor.visit( *this ); }
ExpNode* OrNode::clone(bool copyChildren){return new OrNode(*this);  };
ExpNode::OutType OrNode::getOtype() const { return BOOL; }
ExpNode::Type OrNode::getType() const { return OR; }

string OrNode::get_yicesstr() const {
	stringstream s;
	s << "(" << mother->get_yicesstr() << " || " << father->get_yicesstr()
			<< ")";
	return s.str();
}

void OrNode::addVars(set<int>& varset) {
	mother->addVars(varset);
	father->addVars(varset);
}


/**
 * PlusNode
 */
PlusNode::PlusNode() {
	type = PLUS;
	__is_minus = false;
}

PlusNode::PlusNode(bool minus) {
	__is_minus = minus;
}
PlusNode::PlusNode(const PlusNode& bn) :
	ExpNode(bn), __is_minus(bn.__is_minus) {
}
void PlusNode::set_minus(bool minus) {
	__is_minus = minus;
}
void PlusNode::accept(NodeVisitor& visitor) {
	visitor.visit(*this);
}
ExpNode* PlusNode::clone(bool copyChildren) {
	return new PlusNode(*this);
}
;
ExpNode::OutType PlusNode::getOtype() const {
	return INT;
}
ExpNode::Type PlusNode::getType() const {
	return PLUS;
}

string PlusNode::get_yicesstr() const {
	stringstream s;
	s << "(" << mother->get_yicesstr() << " + " << father->get_yicesstr()
			<< ")";
	return s.str();
}

void PlusNode::addVars(set<int>& varset) {
	mother->addVars(varset);
	father->addVars(varset);
}

/**
 * TimesNode
 */
TimesNode::TimesNode() {
	type = TIMES;
}
TimesNode::TimesNode(const TimesNode& bn) :
	ExpNode(bn) {
}
void TimesNode::accept(NodeVisitor& visitor) {
	visitor.visit(*this);
}
ExpNode* TimesNode::clone(bool copyChildren = true) {
	return new TimesNode(*this);
}

ExpNode::OutType TimesNode::getOtype() const {
	return INT;
}
ExpNode::Type TimesNode::getType() const {
	return TIMES;
}
string TimesNode::get_yicesstr() const {
	stringstream s;
	s << "(" << mother->get_yicesstr() << " * " << father->get_yicesstr()
			<< ")";
	return s.str();
}

void TimesNode::addVars(set<int>& varset) {
	mother->addVars(varset);
	father->addVars(varset);
}

/**
 * DivNode
 */
DivNode::DivNode() {
	type = DIV;
}
DivNode::DivNode(const DivNode& bn) :
	ExpNode(bn) {
}
void DivNode::accept(NodeVisitor& visitor) {
	visitor.visit(*this);
}
ExpNode* DivNode::clone(bool copyChildren = true) {
	return new DivNode(*this);
}
;
ExpNode::OutType DivNode::getOtype() const {
	return INT;
}
ExpNode::Type DivNode::getType() const {
	return DIV;
}
string DivNode::get_yicesstr() const {
	stringstream s;
	s << "(" << mother->get_yicesstr() << " / " << father->get_yicesstr()
			<< ")";
	return s.str();
}

void DivNode::addVars(set<int>& varset) {
	mother->addVars(varset);
	father->addVars(varset);
}

/**
 * GtNode
 */
GtNode::GtNode() {
	type = GT;
}
GtNode::GtNode(const GtNode& bn): ExpNode(bn) {}
void GtNode::accept(NodeVisitor& visitor) {
	visitor.visit(*this);
}
ExpNode* GtNode::clone(bool copyChildren = true) {
	return new GtNode(*this);
}
;
ExpNode::OutType GtNode::getOtype() const {
	return BOOL;
}
ExpNode::Type GtNode::getType() const {
	return GT;
}
string GtNode::get_yicesstr() const {
	stringstream s;
	s << "( > " << mother->get_yicesstr() << " " << father->get_yicesstr() << ")";
	return s.str();
}

void GtNode::addVars(set<int>& varset) {
	mother->addVars(varset);
	father->addVars(varset);
}

/**
 * GeNode
 */

GeNode::GeNode() {
	type = GE;
}
GeNode::GeNode(const GeNode& bn): ExpNode(bn) {}
void GeNode::accept(NodeVisitor& visitor) {
	visitor.visit(*this);
}
ExpNode* GeNode::clone(bool copyChildren = true) {
	return new GeNode(*this);
}
;
ExpNode::OutType GeNode::getOtype() const {
	return BOOL;
}
ExpNode::Type GeNode::getType() const {
	return GE;
}
string GeNode::get_yicesstr() const {
	stringstream s;
	s << "(" << mother->get_yicesstr() << " >= " << father->get_yicesstr()
			<< ")";
	return s.str();
}

void GeNode::addVars(set<int>& varset) {
	mother->addVars(varset);
	father->addVars(varset);
}

/**
 * LtNode
 */

LtNode::LtNode() {type = LT;}
LtNode::LtNode(const LtNode& bn): ExpNode(bn) {}
void LtNode::accept(NodeVisitor& visitor) {
	visitor.visit(*this);
}
ExpNode* LtNode::clone(bool copyChildren = true) {
	return new LtNode(*this);
}
;
ExpNode::OutType LtNode::getOtype() const {
	return BOOL;
}
ExpNode::Type LtNode::getType() const {
	return LT;
}
string LtNode::get_yicesstr() const {
	stringstream s;
	s << "(" << mother->get_yicesstr() << " < " << father->get_yicesstr()
			<< ")";
	return s.str();
}
void LtNode::addVars(set<int>& varset) {
	mother->addVars(varset);
	father->addVars(varset);
}

/**
 * LeNode
 */
LeNode::LeNode() {type = LE;}
LeNode::LeNode(const LeNode& bn): ExpNode(bn) {}
void LeNode::accept(NodeVisitor& visitor) {
	visitor.visit(*this);
}
ExpNode* LeNode::clone(bool copyChildren = true) {
	return new LeNode(*this);
}
;
ExpNode::OutType LeNode::getOtype() const {
	return BOOL;
}
ExpNode::Type LeNode::getType() const {
	return LE;
}
string LeNode::get_yicesstr() const {
	stringstream s;
	s << "(" << mother->get_yicesstr() << " <= " << father->get_yicesstr()
			<< ")";
	return s.str();
}
void LeNode::addVars(set<int>& varset) {
	mother->addVars(varset);
	father->addVars(varset);
}

/**
 * EqNode
 */
EqNode::EqNode() {type = EQ;}
EqNode::EqNode(const EqNode& bn): ExpNode(bn) {}
void EqNode::accept(NodeVisitor& visitor) {
	visitor.visit(*this);
}
ExpNode* EqNode::clone(bool copyChildren = true) {
	return new EqNode(*this);
}
;
ExpNode::OutType EqNode::getOtype() const {
	return BOOL;
}
ExpNode::Type EqNode::getType() const {
	return EQ;
}
string EqNode::get_yicesstr() const {
	stringstream s;
	s << "(" << mother->get_yicesstr() << " == " << father->get_yicesstr()
			<< ")";
	return s.str();
}

void EqNode::addVars(set<int>& varset) {
	mother->addVars(varset);
	father->addVars(varset);
}

/**
 * NeqNode
 */
NeqNode::NeqNode() {type = EQ;}
NeqNode::NeqNode(const NeqNode& bn): ExpNode(bn) {}
void NeqNode::accept(NodeVisitor& visitor) {
	visitor.visit(*this);
}
ExpNode* NeqNode::clone(bool copyChildren = true) {
	return new NeqNode(*this);
}
;
ExpNode::OutType NeqNode::getOtype() const {
	return BOOL;
}
ExpNode::Type NeqNode::getType() const {
	return EQ;
}
string NeqNode::get_yicesstr() const {
	stringstream s;
	s << "(" << mother->get_yicesstr() << " != " << father->get_yicesstr()
			<< ")";
	return s.str();
}

void NeqNode::addVars(set<int>& varset) {
	mother->addVars(varset);
	father->addVars(varset);
}
