#include "ExpressionNodes.h"
#include "ExpDAG.h"

#include <fstream>
#include <sstream>
#include <algorithm>

/**
 * ConstNode
 */
ConstNode::ConstNode() : ExpNode() {
  type = CONST; val = -1;
}
ConstNode::ConstNode(int n) {
  type = CONST; val = n;}
  ConstNode::ConstNode(const ConstNode& bn): ExpNode(bn), val(bn.val){
    //if(val == 13){ cout<<" surprise"<<endl; }
  }
void ConstNode::accept(NodeVisitor& visitor)  { visitor.visit( *this ); }
void ConstNode::setVal(int v){ val = v; }
int ConstNode::getVal() const { return val; }

string ConstNode::get_namestr() const{
  stringstream str;
  str<<__name<<"_C";
  if( val<0){
    str<<"m";
  }
  str<<abs(val);
  return str.str();
}

string ConstNode::get_yicesstr() const{
  stringstream str;
  str << val;
  return str.str();
}

string ConstNode::lid(){
  stringstream str;
  str<<"(" << val << ")";
  return str.str();
}
string ConstNode::lprint(){
  stringstream str;
  str<<id<<"= "<<"(" << val << ")";
  return str.str();
}
ExpNode* ConstNode::clone(bool copyChildren) {
  return new ConstNode(*this);
};
ExpNode::OutType ConstNode::getOtype()const {
  if(otype != BOTTOM){
    return otype;
  }
  /* TODO: Umm??? */
  if(val != 0 && val != 1){
    otype = INT;
  }else{
    otype = BOOL;
  }
  return otype;
}
ExpNode::Type ConstNode::getType() const { return CONST; }

/**
 * BoolNode
 */
BoolNode::BoolNode() : ExpNode() {
  type = BOOLCONST; __val = false;
}
BoolNode::BoolNode(bool n) {
  type = BOOLCONST; __val = n;
}
BoolNode::BoolNode(const BoolNode& bn): ExpNode(bn), __val(bn.__val){
  //if(val == 13){ cout<<" surprise"<<endl; }
}
void BoolNode::accept(NodeVisitor& visitor)  { visitor.visit( *this ); }
void BoolNode::setVal(bool v){ __val = v; }
int BoolNode::getVal() const { return __val; }

string BoolNode::get_namestr() const{
  stringstream str;
  str<<__name<<"_C";
  return str.str();
}

string BoolNode::get_yicesstr() const{
  stringstream str;
  if (__val) {
    str << "true";
  } else {
    str << "false"; 
  }
  return str.str();
}

string BoolNode::lid(){
  stringstream str;
//  str<<"(" << val << ")";
  return str.str();
}
string BoolNode::lprint(){
  stringstream str;
  str<<id<<"= "<<"(" << __val << ")";
  return str.str();
}
ExpNode* BoolNode::clone(bool copyChildren) {
  return new BoolNode(*this);
};
ExpNode::OutType BoolNode::getOtype()const {
  if(otype != BOTTOM){
    return otype;
  }
  return BOOL;
}

ExpNode::Type BoolNode::getType() const { return BOOLCONST; }

/**
 * RecordNode
 */
RecordNode::RecordNode() {
  type = RECORD;
  /* TODO */
  __record = new vector<eid_t>();
  __num_fields = 0;
}
RecordNode::~RecordNode() {
  delete __record;
}
RecordNode::RecordNode(int numfields) {
#ifdef __DEBUG_1
	cout << "RecordNode::RecordNode: Creating record with " << numfields
		 << " fields." << endl;
#endif
	type = RECORD;
	__record = new vector<eid_t>(numfields, -1);
	__num_fields = numfields;
	cout << "done allocating" << endl;
}
RecordNode::RecordNode(const RecordNode& bn)
  : ExpNode(bn), __record(bn.__record), __num_fields(bn.__num_fields) { }

void RecordNode::accept(NodeVisitor& visitor)  { visitor.visit( *this ); }
string RecordNode::get_namestr() const{
  stringstream str;
  str<<__name<<"_C";
/*  str<<abs(val); */
  return str.str();
}
string RecordNode::lid(){
  stringstream str;
/*  str<<"(" << val << ")"; */
  return str.str();
}
string RecordNode::lprint(){
  stringstream str;
  vector<eid_t>::iterator record_it;
  for (record_it = __record->begin(); record_it != __record->end(); ++record_it) {
	  str << *record_it << " ";
  }

  return str.str();
}
ExpNode* RecordNode::clone(bool copyChildren) {
  return new RecordNode(*this);
};
ExpNode::OutType RecordNode::getOtype()const {
  if(otype != BOTTOM){
    return otype;
  } else {
    return RECORDTY;
  }
}

ExpNode::Type RecordNode::getType() const { return RECORD; }

eid_t RecordNode::lookupField(int field) {
#ifdef __DEBUG_1
	cout << "RecordNode::lookupField: " << field << " in record of size "
		 << __record->size() << endl;
#endif
  Assert( field < __record->size()
        , "Must access a field in the record.");
  return (*__record)[field];
}

void RecordNode::setValue(int field, eid_t eid) {
#ifdef __DEBUG_1
  cout << "RecordNode:: setValue: " << field << " := " << eid << endl;
#endif
  (*__record)[field] = eid;
#ifdef __TEST
  Assert( __record->size() > field
        , "Record size should be at least as big as the number of fields.");
  int lookup_eid = this->lookupField(field);
  stringstream ss;
  ss << eid << " != " << lookup_eid;
  string neq_str = ss.str();
  Assert( lookup_eid == eid, neq_str);
#endif
}

/**
 * RelationNode
 */
RelationNode::RelationNode() {
  type = RELATION;
  __next_itid = 0;
  __num_fields = 0;
}
RelationNode::~RelationNode() {

}
RelationNode::RelationNode(const RelationNode& bn):
  ExpNode(bn), __rel(bn.__rel), __next_itid(bn.__next_itid), __num_fields(bn.__num_fields) {
  cout << "RelationNode::RelationNode(const RelationNode& bn)" << endl;
}

void Tokenize(const string& str, vector<string>& tokens,
		const string& delimiters = " ") {
	// Skip delimiters at beginning.
	string::size_type lastPos = str.find_first_not_of(delimiters, 0);
	// Find first "non-delimiter".
	string::size_type pos = str.find_first_of(delimiters, lastPos);

	while (string::npos != pos || string::npos != lastPos) {
		// Found a token, add it to the vector.
		tokens.push_back(str.substr(lastPos, pos - lastPos));
		// Skip delimiters.  Note the "not_of"
		lastPos = str.find_first_not_of(delimiters, pos);
		// Find next "non-delimiter"
		pos = str.find_first_of(delimiters, lastPos);
	}
}

/* Construct a relation from file. */
/* TODO: Do something with numfields. */
RelationNode::RelationNode(int numfields, const char* file_name) {
#ifdef __DEBUG_1
	cout << "RelationNode::RelationNode: Reading from file " << file_name
			<< endl;
#endif
	__num_fields = numfields;

	ifstream file;
	file.open(file_name);
	Assert(file.is_open(), "Was not able to open file.");

	int pkey = 0;
	while (!file.eof()) {
		/* Get current line. */
		string curline;
		getline(file, curline);
#ifdef __DEBUG_1
		cout << "RelationNode::RelationNode: Read line " << curline << endl;
#endif

		/* Scan current line. */
		vector<string> tokens;
		Tokenize(curline, tokens);
		if (tokens.size() == __num_fields) {
			RecordNode* curRecord = new RecordNode(__num_fields);
			__rel[pkey] = curRecord;

			vector<string>::iterator token_it;
			int field = 0;
			for (token_it = tokens.begin(); token_it != tokens.end(); ++token_it) {
				bool is_digit = true;

				string::iterator pch;
				for (pch = token_it->begin(); pch != token_it->end(); ++pch) {
					if (!isdigit(*pch)) {
						is_digit = false;
					}
				}

#ifdef __TEST
				Assert(curRecord != NULL, "Should have a primary key.");
#endif
				if (is_digit) {
					int cur_val = atoi(token_it->c_str());
					ConstNode* cnode = new ConstNode(cur_val);
					__rel[pkey]->setValue(field, cnode->get_name());
				} else {
					/* Otherwise insert symbolic value. */
					SymNode* cnode = new SymNode();
					__rel[pkey]->setValue(field, cnode->get_name());
				}
				++field;
			} // rof token_it
#ifdef __DEBUG_1
			cout << "RelationNode::RelationNode:: Read record "
				 << curRecord->get_name() << ": " << curRecord->lprint() << endl;
#endif
		}

		++pkey; /* Increment the primary key. */
	} // elihw reading file
	file.close();
}
RelationNode::RelationNode(int numfields) {
	  type = RELATION;
	  __next_itid = 0;
	  __num_fields = numfields;
}

void RelationNode::accept(NodeVisitor& visitor)  { visitor.visit( *this ); }
string RelationNode::get_namestr() const{
  stringstream str;
  str<<__name<<"_C";
  /* if( val<0){
    str<<"m";
  } */
  // str<<abs(val);
  return str.str();
}
string RelationNode::lid(){
  stringstream str;
  // str<<"(" << val << ")";
  return str.str();
}
string RelationNode::lprint(){
  stringstream str;
  map<int, RecordNode*>::iterator rel_it;
  for (rel_it = __rel.begin(); rel_it != __rel.end(); ++rel_it) {
	  str << (rel_it->second)->lprint() << ", ";
  }
  // str<<id<<"= "<<"(" << val << ")";
  return str.str();
}
ExpNode* RelationNode::clone(bool copyChildren) {
  return new RelationNode(*this);
};
ExpNode::OutType RelationNode::getOtype()const {
  if(otype != BOTTOM){
    return otype;
  } else {
    return RELTY;
  }
}

ExpNode::Type RelationNode::getType() const { return RELATION; }

map<int, RecordNode*>& RelationNode::get_relation() {
	return __rel;
}

int RelationNode::getIterator() {
  int id = __next_itid;
  __iterators[__next_itid++] = __rel.begin();
#ifdef __DEBUG_1
  cout << "RelationNode::getIterator: Got iterator with ID " << id
       << ", should have " << __rel.size() << " elements." << endl;
#endif
  return id;
}

eid_t RelationNode::nextRecord(int itid) {
  relation_it cur_it = __iterators[itid];
#ifdef __DEBUG_1
  cout << "RelationNode::nextRecord: next record for " << itid << endl;
  Assert(__iterators.find(itid) != __iterators.end()
        , "Iterator exists.")
#endif
  if (cur_it == __rel.end()) {
    return -1;
  }
  eid_t cur_eid = (cur_it->second)->get_name();
  __iterators[itid] = ++cur_it;
#ifdef __DEBUG_1
  cout << "RelationNode::nextRecord: next record has ID " << cur_eid << endl;
#endif
  return cur_eid;
}

eid_t RelationNode::lookupRelationVal(int pkey, int field) {
  if (__rel.find(pkey) != __rel.end()) {
    Assert(__rel[pkey] != NULL, "Record in relation cannot be null.");
    Assert(__rel[pkey]->getType() == RECORD
          , "Relation should contain records.");
    printf("I get here.\n");
    return __rel[pkey]->lookupField(field);
  } else {
    return -1;
  }
}

void RelationNode::setRecord(int pkey, RecordNode* record) {
  Assert(record != NULL, "New record cannot be null.");
  __rel[pkey] = record;
}

void RelationNode::setValue(int pkey, int field, eid_t eid) {
#ifdef __DEBUG_1
  cout << "RelationNode::setValue: " << pkey << ", " << field << " := " << eid
       << endl;
#endif

#ifdef __TEST
  Assert(__rel.find(pkey) != __rel.end(), "pkey should exist in relation");
  Assert(__rel[pkey] != NULL
        , "relation corresponding to primary key should not be null");
#endif
 
  __rel[pkey]->setValue(field, eid);

#ifdef __DEBUG_1
  cout << "RelationNode::setValue: Finished setting value for " << pkey << endl;
#endif
}

int RelationNode::get_num_fields() { return __num_fields; }
