#include <stdio.h>
#include <caml/mlvalues.h>
#include <caml/memory.h>
#include <assert.h>
/* TODO: Combine these two files. */
#include "BackendRep/backend_defs.h"

/**
 * Declarations.
 */
void createArithopNode(arithop_t op);
void createBoolopNode(boolop_t op);
void bindBackendVal(int varname);
void lookupBackendVal(int varname);
void lookupBackendValHidden(int varname);
/* void lookupValAsExp(eid_t eid, int* type, int* val); */
void concretizeEnvVal(eid_t eid, int val);

eid_t createConstNode(int i);
eid_t createSymNode(int varname);
eid_t createStringNode(const char* s);

int expIsSym();

void printEnvVars();
void printEnvExp();

void initBackendEnv();
void clearBackendEnv();
void printBackendLive();

/**
 * Returns the EID of a constant, symbolic value, or EID.
 */
CAMLprim value getAtomEid(value exp) {
  int id = -1;
  
  assert (Is_block(exp));
  if (Tag_val(exp) == 0) /* Int */ {
    id = createConstNode(Int_val(Field(exp, 0)));
  } else if (Tag_val(exp) == 1) /* Sym */ {
    id = createSymNode(Int_val(Field(exp, 0)));
  } else if (Tag_val(exp) == 2) /* Bool */ {
    id = createBoolNode(Bool_val(Field(exp, 0)));
  } else if (Tag_val(exp) == 3) /* String */ {
    id = createStringNode(String_val(Field(exp, 0)));
  }
  /* TODO: Guarded expressions? */
  return Val_int(id);
}

CAMLprim void getCondEid() {
  createCondNode();
}

/**
 * This function takes a basic arithmetic operation and its two arguments and
 * stores it in the backend.
 */
CAMLprim void basicArithop(value op) {
  createArithopNode(Int_val(op));
}

CAMLprim value basicBoolop(value op) {
  createBoolopNode(Int_val(op));
}

/**
 * This function returns the value associated with a particular expression ID.
 *
 * TODO: Do we need this function?
 */
CAMLprim value eidValue(value eid) {
  CAMLparam0();
  int type, val;
  lookupValAsExp(Int_val(eid), &type, &val);

  int tag;
  switch(type) {
    case 8:
      tag = 0;
      break;
    case 10:
      tag = 1;
      break;
    case 9:
      tag = 3;
      break;
    default:
      tag = 6;
      val = -1;
  }

  CAMLlocal1(result);
  result = caml_alloc(1, tag);
  Store_field(result, 0, Val_int(val));
  CAMLreturn (result);
}

/**
 * This function calls the appropriate backend function that associates the
 * given variable name with the given expression ID.
 */
CAMLprim void bindVal(value varname) {
  bindBackendVal(Int_val(varname));
}

/**
 * This function calls the appropriate backend function to get the expression
 * ID associated with a variable name.
 */
CAMLprim void lookupVal(value varname) {
  lookupBackendVal(Int_val(varname));
}

CAMLprim void lookupValHidden(value varname) {
  lookupBackendValHidden(Int_val(varname));
}

CAMLprim value getConcreteBool() {
  int boolval = getEnvConcreteBool();
  if (boolval >= 0) {
    return Val_bool(boolval);
  } else {
    caml_failwith("Did not get a concrete value.");
  }
}

CAMLprim value getConcreteInt() {
  int concrete = 0;
  int intval = getEnvConcreteInt(&concrete);
  if (concrete) {
    return Val_int(intval);
  } else {
    caml_failwith("Did not get a concrete value.");
  }
}

CAMLprim void concretizeSym(value vname, value val) {
  concretizeEnvVal(Int_val(vname), Int_val(val));
}

CAMLprim value isSym() {
  return Val_bool(expIsSym());
}

CAMLprim void printEnv() {
  printEnvVars();
}

CAMLprim value printExp() {
  printEnvExp();
}

void initEnv() { initBackendEnv(); }
void clearEnv() { clearBackendEnv(); }
void printLive() { printBackendLive(); }
