module BackendInterface
  : sig
    type rit = int
    type arithop = int
    type boolop = int

    type bexp =
      | Int of int
      | Sym of int
      | Bool of bool
      | String of string
      | GuardedExp of bexp        (* Guard *)
        * bexp * bexp (* True case, false case *)
      | Nothing

    (* Backend stack manipulation functions. *)
    val pop : unit -> unit
    val dup : unit -> unit
    val swap : unit -> unit
    val pushUnit : unit -> unit

    (* Takes a file name and returns a record ID. *)
    val readRelation : int -> string -> unit
    val mkRelation : int -> int -> unit (* Give it varname and num fields *)

    val bindRelationVal : int -> int -> unit
    val lookupRelationVal : int -> int -> unit
    val lookupFieldIndex : int -> int -> unit
    val lookupFieldIndexHidden : int -> int -> unit

    (* TODO: Do we also want a separate createRelation function? *)
    val relIterator : unit -> unit
    val relPkeyIterator : int -> unit
    val nextRecord : unit -> bool
    val recordSelect : int -> rit
    val listElem : unit -> unit

    (***************************)
    (* Evaluating expressions. *)
    (***************************)
    val basicArithop : arithop -> unit
    val basicBoolop : boolop -> unit
    val createAtomEid : bexp -> unit
    val createCondEid : unit -> unit

    val bindVal : int -> unit
    val lookupVal : int -> unit
    val lookupValHidden  : int -> unit

    val eqExps : unit -> bool
    val eqExpToVal : int -> bool

    val isSym : unit -> bool
    val getConcreteBool : unit -> bool
    val getConcreteInt : unit -> int
  
    val concretizeSym : int -> int -> unit

    (***********************)
    (* Adding constraints. *)
    (***********************)
    val assertExp : unit -> unit
    val writeAssertions : unit -> unit
    val resolveWithContext : int -> unit

    (************)
    (* Printing *)
    (************)
    val printEnv : unit -> unit
    val printExp : unit -> unit

    val initEnv : unit -> unit
    val clearEnv : unit -> unit
    val printLive : unit -> unit
  end
