(*
  Marriage inference example.
  Last updated: 11/11/09
 *)
open AstUtils
open Error
open Evalenv

module Ast = Ast.Ast
module Evaluator = Evaluator.Evaluator
module TestUtils = TestUtils.TestUtils

let household = 0
let line = 1
let age = 2
let married = 3
let spouse = 4
let hot_lunch = 5
let census_fields =
  [ (household, Ast.IntType)  (* household *)
  ; (line,      Ast.IntType)  (* field of interest *)
  ; (age,       Ast.IntType)
  ; (married,   Ast.IntType)
  ; (spouse,    Ast.IntType)
  ; (hot_lunch, Ast.IntType)
  ]

let data_ty_name = TestUtils.gen_var ()
let data_ty =
  { Ast.dataname = data_ty_name
  ; Ast.fieldinfo = census_fields }
let data_ty_imputed_name = TestUtils.gen_var ()
(*
  data data_ty { ... }
  where (rel) {
    forall e in rel .
      forall e' in rel.
        (e.household == e'.household &&
          e.spouse == e'.line)==>
          (e'.spouse == e.line &&
            e.married == e.married)
  }
 *)

(* Define a data type. *)
(*
 *)
(* Variables *)
let in_data_missing_name = TestUtils.gen_var ()
let in_data_whole_name = TestUtils.gen_var ()
let rel = TestUtils.gen_var ()
let bound_constraint =
  let e = TestUtils.gen_var () in
  let e' = TestUtils.gen_var () in
    let spouse_match_expr =
    let line_match =
      TestUtils.mkACBoolop Ast.Equiv
        (ref (Ast.VarExp (Ast.FieldIndexVar (e, spouse))))
        (ref (Ast.VarExp (Ast.FieldIndexVar (e', line)))) in
      TestUtils.mkACBoolop Ast.And
        (TestUtils.mkACBoolop Ast.Nequiv
            (ref (Ast.VarExp (Ast.FieldIndexVar (e, line))))
            (ref (Ast.VarExp (Ast.FieldIndexVar (e', line)))))
        line_match in
  let line_no_same =
    TestUtils.mkACBoolop Ast.Equiv
      (ref (Ast.VarExp (Ast.FieldIndexVar (e, line))))
      (ref (Ast.VarExp (Ast.FieldIndexVar (e', line)))) in
  let age_constraint =
    TestUtils.mkACBoolop Ast.Equiv
      (ref (Ast.VarExp (Ast.FieldIndexVar (e, age))))
      (ref
         (Ast.Arithop
            ( Ast.Plus
            , ref (Ast.PrimExp (Ast.Integer 1))
            , ref (Ast.VarExp (Ast.FieldIndexVar (e', age)))))) in
  let hot_lunch_constraint =
    TestUtils.mkACBoolop Ast.Equiv
      (ref (Ast.VarExp (Ast.FieldIndexVar (e, hot_lunch))))
      (ref (Ast.VarExp (Ast.FieldIndexVar (e', hot_lunch)))) in
  let line_match =
    TestUtils.mkACBoolop Ast.Equiv
      (ref (Ast.VarExp (Ast.FieldIndexVar (e, spouse))))
      (ref (Ast.VarExp (Ast.FieldIndexVar (e', line)))) in
  let married_same_as_spouse =
    TestUtils.mkACBoolop Ast.Equiv
      (ref (Ast.VarExp (Ast.FieldIndexVar (e, married))))
      (ref (Ast.VarExp (Ast.FieldIndexVar (e', married)))) in
  let spouse_symmetric =
    TestUtils.mkACBoolop Ast.Equiv
      (ref (Ast.VarExp (Ast.FieldIndexVar (e', spouse))))
      (ref (Ast.VarExp (Ast.FieldIndexVar (e, line)))) in
  let married_constraint =
    TestUtils.mkACBoolop Ast.And married_same_as_spouse spouse_symmetric in
      [ Ast.ConstraintForMatch
             ( e, rel
               , e', rel
               , spouse_match_expr
               , married_constraint)
      ;  Ast.ConstraintExistsMatch
              ( e, rel
                , e', in_data_whole_name
                , Ast.FieldIndexVar (e', age)
                , line_no_same
                , age_constraint)
      ; Ast.ConstraintExistsMatch
              ( e, rel
                , e', rel
                , Ast.FieldIndexVar (e, hot_lunch)
                , Ast.ref_true
                , hot_lunch_constraint) ]
let tydecl = Ast.TypeDecl (data_ty, rel, bound_constraint)

(* Declare a relation. *)
  let data file =
    Ast.FunAppExp
      ( -1
        , [ ref (Ast.PrimExp (Ast.String file)) ] )
  (* TODO: Directly call the input handler on generated data here. *)
  (* Ast.PrimExp
                (Ast.RelationExp
                    ( indata_name
                    , data_ty_name
                    , TestUtils.list2relation indata_name input_list )) *)
  let assign_in file data_name tyname =
    Ast.AssignDecl ( data_name
                  , Ast.RelationType tyname
                  , Ast.RhsExp (data file) )

  (*
  * Declare a function that sums over things.
  *
  * procedure sum_helper (entry : census_data, sum : int) =
  *  if entry.married
  *    then sum + 1
  *    else sum
  *
  * procedure sum_ages ( r : census_data relation ):
  *  let sum_age = fold e cond_sum_married 0 r in
  *    sum_age
  *)
  let sum_helper = TestUtils.gen_var ()
  let sum_helper_fun =
    let sum = TestUtils.gen_var () in
    let census_entry = TestUtils.gen_var () in
    let married_cond =
      ref (Ast.Boolop
             ( Ast.Equiv
             , ref (Ast.VarExp (Ast.FieldIndexVar (census_entry, married)))
             , ref (Ast.PrimExp (Ast.Integer 1)))) in
    let hotlunch_cond =
      ref (Ast.Boolop
             ( Ast.Gt
             , ref (Ast.VarExp (Ast.FieldIndexVar (census_entry, hot_lunch)))
             , ref (Ast.PrimExp (Ast.Integer 0)))) in
    Ast.FunDecl
      ( (sum_helper, Ast.IntType)
      , { Ast.funargs = [ (census_entry, Ast.RecordType data_ty_name)
                        ; (sum, Ast.IntType) ]
        ; Ast.funbody =
            ref
              (Ast.CondExp
               ( ref (Ast.Boolop (Ast.And, married_cond, hotlunch_cond))
               , TestUtils.mkArithop Ast.Plus
                (ref (Ast.VarExp (Ast.Var sum)))
                (ref (Ast.VarExp (Ast.FieldIndexVar (census_entry, age))))
               , ref (Ast.VarExp (Ast.Var sum))))
        } )
  (* TODO: test let *)
  let sum_ages = TestUtils.gen_var ()
  let sum_ages_fun =
    let r = TestUtils.gen_var () in
    Ast.FunDecl
      ( (sum_ages, Ast.IntType)
      , { Ast.funargs = [ (r, Ast.RelationType data_ty_name) ]
        ; Ast.funbody =
          ref (Ast.FoldExp ( sum_helper
                      , ref (Ast.PrimExp (Ast.Integer 0))
                      , r )) } )

(* Call the function we just made. *)
let sum_data_name = TestUtils.gen_var ()
let sum_data =
  Ast.AssignDecl
    ( sum_data_name, Ast.IntType
    , Ast.RhsExp
        (Ast.FunAppExp
           (sum_ages, [ ref (Ast.VarExp (Ast.Var in_data_missing_name))])) )

(* Program. *)
let program files =
  [ tydecl
  ; assign_in (List.nth files 1) in_data_whole_name data_ty_name
  ; assign_in (List.nth files 0) in_data_missing_name data_ty_name
  ; sum_helper_fun
  ; sum_ages_fun
  ; sum_data
  ; TestUtils.mkConcreteResult TestUtils.result_name Ast.IntType sum_data_name ]
