(*
 *)
open AstUtils
open Error
open Evalenv

module Ast = Ast.Ast
module Evaluator = Evaluator.Evaluator
module TestUtils = TestUtils.TestUtils

(*
 * data census_data_cleaned { ... }
 * where {
 *  record.married ==> record.spouse == relation.name(record.spouse).spouse.
 *  record.married ==> relation.name(record.spouse).married.
 *  }
 *)

let mkConcreteResult ty v =
  Ast.AssignDecl ("result", ty, Ast.Concretize (Ast.VarExp (Ast.Var v)))

(* Define a data type. *)
let censusty_name = "censusty"
let bound_constraint =
  let v = "v" in
  let get_spouse_married = Ast.ACPrim 0 in
    (*
    Ast.ACMatch
      ( Ast.Var "rel"
      , "e'"
      , Ast.ACBoolop
          ( Ast.Equiv
          , Ast.ACVarExp (Ast.FieldIndexVar ("e", "spouse"))
          , Ast.ACVarExp (Ast.FieldIndexVar ("e'", "name")) )
      , "married" ) in *)
  let married_same_as_spouse =
    Ast.ACBoolop ( Ast.Equiv
                 , Ast.ACVarExp (Ast.FieldIndexVar ("e", "married"))
                 , get_spouse_married) in
  let get_spouse_spouse = Ast.ACPrim 0 in
    (*
     Ast.ACMatch
      ( Ast.Var "rel"
      , "e'"
      , Ast.ACBoolop
          ( Ast.Equiv
          , Ast.ACVarExp (Ast.FieldIndexVar ("e", "spouse"))
          , Ast.ACVarExp (Ast.FieldIndexVar ("e'", "name")) )
      , "spouse" ) in *)
  let spouse_symmetric =
    Ast.ACBoolop ( Ast.Equiv
                 , Ast.ACVarExp (Ast.FieldIndexVar ("e", "spouse"))
                 , get_spouse_spouse) in
  let married_constraint =
    Ast.ACBoolop ( Ast.And
                 , married_same_as_spouse
                 , spouse_symmetric ) in
  Ast.ConstraintAssume
    [( "rel"
     , Ast.ConstraintForall ("e"
                          , Ast.Var "rel"
                          , Ast.ACTrue
                          , married_constraint) )]
let tydecl = Ast.TypeDecl (Censusdata.censusty, [bound_constraint])

(* Declare a relation. *)
(* TODO: Change this to involve reading input. *)
let input_list =
  let name = "name" in
  let age = "age" in
  let m = "married" in
  let s = "spouse" in
  let c = "child" in
    [ (* parent *)
      (0, [(name, "0"); (age, "42"); (m, "??"); (s, "??"); (c, "-1") ])
      (* child *)
    ; (1, [(name, "1"); (age, "44"); (m, "1"); (s, "0"); (c, "-1") ]) ]
let indata_name = "in_rel"
let data = Ast.PrimExp
               (Ast.RelationExp
                  ( indata_name
                  , censusty_name
                  , ref (TestUtils.list2relation indata_name input_list) ))
let assign_in = Ast.AssignDecl ( indata_name
                               , Ast.RelationType censusty_name
                               , Ast.RhsExp data )

(*
 * Declare a function that sums over things.
 *
 * procedure sum_helper (census_entry : census_data, sum : int) =
 *  if (census_entry.married == 1)
 *    then sum + census_entry.age
 *    else sum
 *
 * procedure sum_ages ( r : census_data relation ):
 *  fold e cond_sum_married 0 r
 *)
let sum_helper_fun =
  let sum_exp =
    Ast.Binop
      ( Ast.Plus
      , Ast.VarExp (Ast.Var "sum")
      , Ast.VarExp (Ast.FieldIndexVar ("census_entry", "age"))) in
  Ast.FunDecl
    { Ast.funname = "sum_helper"
    ; Ast.funargs = [ ("census_entry", Ast.RecordType censusty_name)
                    ; ("sum", Ast.IntType) ]
    ; Ast.funbody =
      Ast.CondExp
        ( Ast.Boolop
            ( Ast.Equiv
            , Ast.VarExp (Ast.FieldIndexVar ("census_entry", "married"))
            , Ast.PrimExp (Ast.Integer 1))
        , sum_exp, Ast.VarExp (Ast.Var "sum"))
    }
(* TODO: test let *)
let sum_ages_fun =
  Ast.FunDecl
    { Ast.funname = "sum_ages"
    ; Ast.funargs = [ ("r", Ast.RelationType censusty_name) ]
    ; Ast.funbody =
        Ast.FoldExp ( "sum_helper"
                    , (Ast.PrimExp (Ast.Integer 0), false)
                    , Ast.VarExp (Ast.Var "r") ) }

(* Call the function we just made. *)
let sum_data =
  Ast.AssignDecl
    ( "sum_data", Ast.IntType
    , Ast.RhsExp
        (Ast.FunAppExp ("sum_ages", [Ast.VarExp (Ast.Var indata_name)])) )

(* Program. *)
let program = [ tydecl
              ; assign_in
              ; sum_helper_fun
              ; sum_ages_fun
              ; sum_data
              ; TestUtils.mkConcreteResult
                  "result" Ast.IntType "sum_data" ]
