(*
 *)
open AstUtils
open Error
open Evalenv

module Ast = Ast.Ast
module Evaluator = Evaluator.Evaluator
module TestUtils = TestUtils.TestUtils

(*
  data census_data_cleaned { ... }
  where {
    relation.name(c).age <= record.age - 12.
  }
 *)

let mkConcreteResult ty v =
  Ast.AssignDecl ("result", ty, Ast.Concretize (Ast.VarExp (Ast.Var v)))

(* Define a data type. *)
let censusty_name = "censusty"
let bound_constraint =
  let v = "v" in
  let get_child_age = Ast.ACPrim 0 in
    (*
    Ast.ACMatch
      ( Ast.Var "rel"
      , "e'"
      , Ast.ACBoolop
          ( Ast.Equiv
          , Ast.ACVarExp (Ast.FieldIndexVar ("e'", "name"))
          , Ast.ACVarExp (Ast.FieldIndexVar ("e", "child")) ) 
      , "age" ) in *)
  let incc =
    Ast.ACBoolop ( Ast.Gt
                 , Ast.ACVarExp (Ast.FieldIndexVar ("e", "age"))
                 , Ast.ACArithop
                       (Ast.Plus, get_child_age, Ast.ACPrim 12)) in
  Ast.ConstraintAssume
    [( "rel"
     , Ast.ConstraintForall ( "e"
                          , Ast.Var "rel"
                          , Ast.ACTrue 
                          , incc) )]
let tydecl = Ast.TypeDecl (Censusdata.censusty, [bound_constraint])

(* Declare a relation. *)
(*
let input_list =
  let name = "name" in
  let age = "age" in
  let m = "married" in
  let s = "spouse" in
  let c = "child" in
    [ (* parent *)
      (0, [(name, "0"); (age, "??"); (m, "1"); (s, "3"); (c, "1") ])
      (* child *)
    ; (1, [(name, "1"); (age, "5"); (m, "0"); (s, "-1"); (c, "-1") ]) ]
*)
let indata_name = "in_rel"
let data =
  Ast.FunAppExp
    ( "input"
      , [ Ast.PrimExp (Ast.String "data/integerBounds") ] )
(* TODO: Directly call the input handler on generated data here. *)
(* Ast.PrimExp
               (Ast.RelationExp
                  ( indata_name
                  , censusty_name
                  , TestUtils.list2relation indata_name input_list )) *)
let assign_in = Ast.AssignDecl ( indata_name
                               , Ast.RelationType censusty_name
                               , Ast.RhsExp data )

(*
 * Declare a function that sums over things.
 *
 * procedure sum_helper (census_entry : census_data, sum : int) =
 *  sum + census_entry.age
 *
 * procedure sum_ages ( r : census_data relation ):
 *  let sum_age = fold e cond_sum_married 0 r in
 *    sum_age / size r
 *)
let sum_helper_fun =
  Ast.FunDecl
    { Ast.funname = "sum_helper"
    ; Ast.funargs = [ ("census_entry", Ast.RecordType censusty_name)
                    ; ("sum", Ast.IntType) ]
    ; Ast.funbody =
        Ast.Binop
          ( Ast.Plus
          , Ast.VarExp (Ast.Var "sum")
          , Ast.VarExp (Ast.FieldIndexVar ("census_entry", "age")))
    }
(* TODO: test let *)
let sum_ages_fun =
  Ast.FunDecl
    { Ast.funname = "sum_ages"
    ; Ast.funargs = [ ("r", Ast.RelationType censusty_name) ]
    ; Ast.funbody =
        Ast.Binop
          ( Ast.Divide
          , Ast.FoldExp ( "sum_helper"
                        , (Ast.PrimExp (Ast.Integer 0), false)
                        , Ast.VarExp (Ast.Var "r") )
          , Ast.FunAppExp ("size", [Ast.VarExp (Ast.Var "r")] )) }

(* Call the function we just made. *)
let sum_data =
  Ast.AssignDecl
    ( "sum_data", Ast.IntType
    , Ast.RhsExp
        (Ast.FunAppExp ("sum_ages", [Ast.VarExp (Ast.Var indata_name)])) )

(* Program. *)
let program = [ tydecl
              ; assign_in
              ; sum_helper_fun
              ; sum_ages_fun
              ; sum_data
              ; TestUtils.mkConcreteResult
                  "result" Ast.IntType "sum_data" ]
