(*
  Module      : benchmark.ml
*)
open AstUtils
open BackendInterface
open Error
open Evalenv

module Ast = Ast.Ast
module Evaluator = Evaluator.Evaluator
open TestUtils

(* Specifications for command-line arguments. *)
let infer_files = ref []
let set_file file =
  print_endline ("Reading from " ^ file);
  infer_files := file::!infer_files

let run_married = ref false
let run_long = ref false
let run_hot = ref false
let run_combined = ref false
let cmd_specs =
  [ ("-file",     Arg.Rest set_file,     "file" ) 
  ; ("-debug",    Arg.Int set_debug,     "debug?")
  ; ("-married",  Arg.Set run_married,   "run married?")
  ; ("-long",     Arg.Set run_long,      "run longetudinal?")
  ; ("-hot",      Arg.Set run_hot,       "run hot deck?")
  ; ("-combined", Arg.Set run_combined,  "run combined example?") ]
let anon_fun x = print_endline x
let usage_message = "Please use the following options."

let benchmarks () =
  [ ( "Marriage" ,    InferMarried.program,  run_married )
  ; ( "Longitudinal", InferLong.program,     run_long )
  ; ( "Hot deck",     HotDeck.program,       run_hot)
  ; ( "Combined",     CombinedExample.program, run_combined) ]
(*
  [ ("Conditional sum"  , ConditionalSum.program)
  ; ("Integer bounds"   , IntegerBounds.program) ]
 *)

(* Main stuff. *)
let main () =
  let _ = Arg.parse cmd_specs anon_fun usage_message in
  let run_benchmark (cur_prog_name, prog, run) : unit =
    if !run
    then
      (print_endline ("Running " ^ cur_prog_name);
      let cur_prog = prog !infer_files in
      let _ =
        if debug_level 1
        then AstPrint.print_topdecls cur_prog 
        else () in
      let _ = Time.reset_time () in
      let start_time = Sys.time () in
      let denv = ref Evalenv.empty_evalenv in
      let _ = Evaluator.evaluate denv cur_prog in
      let end_time = Sys.time () in
        (* Look up the result in the environment. *)
        Evalenv.lookupValueEnv denv (Ast.Var TestUtils.result_name);
        debug_print 1 ("RESULT FOR " ^ cur_prog_name ^ ":");
        (if debug_level 1
             then BackendInterface.printExp ()
             else ());
        Printf.printf "Running time:\t %f\n" (end_time -. start_time);
        Time.print_time ())
  (* TODO: get time results out of here. *)
  in
  BackendInterface.initEnv ();
  List.iter run_benchmark (benchmarks ());
  BackendInterface.printLive ()

let _ = Printexc.print main ()
