(* Builtins *)
open AstUtils
open Evalenv

module Ast = Ast.Ast

module Builtins =
  struct
    open Error

    (* Build in relation set sorting routine. *)
                         (*
    let sortList (env : Evalenv.evalenv)
                 (result_name : string)
                (args : string list)
              : Evalenv.evalenv =
      (* Next ID's are unique. If we are the next ID of one guy then we can't be
         the  next ID of any other guy. *)
      (*
      let get_unique_constraints n =
        let constraints = ref [] in
        (* Each of i through n could be next. *)
        for i=0 to n do
          for j=0 to (n-1) do
            (* If i is j's next, then i can't be the next of any other one. *)
            let next_pred =
              Constraint.CBoolop
                ( Constraint.CEq, Constraint.CValue (result_name, "nid", j)
                , Constraint.CInt i) in
            let cur_constraints = ref [] in
            for k=0 to (n-1) do
              if k != j
                then
                  let cur_neq_constraint =
                    Constraint.CBoolop
                      ( Constraint.CNeq, Constraint.CValue (result_name, "nid", k)
                      , Constraint.CInt i ) in
                    cur_constraints := cur_neq_constraint::!cur_constraints
                else ()
            done;
            constraints :=
              Constraint.CGuard (next_pred
                , Constraint.CAnd !cur_constraints, Constraint.CTrue)::
                !constraints;
          done
        done;
        !constraints in *)
        (* Look up what we want to sort. *)
        if List.length args == 2
          then
            let data_name = List.nth args 0 in
            let fieldname = List.nth args 1 in
              (* We now have our hands on the previous relation. *)
              (match Evalenv.lookupRelationGlobalEnv env (Ast.Var data_name) with
                Some data_relation ->
                  (* let unique_constraints =
                        get_unique_constraints
                          (MapUtils.int_map_size data_relation) in *)
                  let sorted_records =
                    SortUtils.sortRecords
                      env result_name fieldname data_relation in
                    Evalenv.bindValueGlobalEnv env (Ast.Var result_name)
                          (Ast.PrimExp
                             (Ast.RelationExp (result_name, sorted_records)))
              | None -> raise (Unbound (data_name, "Builtins.sort")))
          (* Sort the records in the current relation. *)
          else raise (Unexpected "Sort on too many arguments.")

    (* Does union of 2 relations.  Doesn't do any sort of duplicate removal
       because everything has a unique ID. *)
    let unionRelations env rel1 rel2 : Ast.relation =
      let appendRelations start2
                          (rel1 : Ast.relation)
                          (rel2 : Ast.relation)
                        : Ast.relation =
        (* Add element of second relation to currently growing relation. *)
        let addRelation cur_uid cur_record cur_relation : Ast.relation =
          let new_uid = cur_uid + start2 in
          let cur_nid_exp =
              match RelationUtils.get_index_field "nid" cur_uid cur_relation with
            | Some e -> e
            | None -> raise (Unexpected "field does not have nid") in
          (*  TODO: add length1 *)
          let new_nid_exp =
            EvalExp.evalExp env
              (Ast.Binop ( Ast.Plus
                        , cur_nid_exp
                        , Ast.PrimExp (Ast.Integer start2))) in
      let record' =
        RelationUtils.set_field "nid" new_nid_exp cur_record in
        IntMap.add new_uid record' cur_relation in
    (* TODO: update "next" ID of the last relation in relation 1 *)
      IntMap.fold addRelation rel1 rel2 in
  
  if debug_level 3
    then Printf.printf "Builtins.unionRelations: %s and %s\n" rel1 rel2
    else ();
  match Evalenv.lookupRelationGlobalEnv env (Ast.Var rel1),
        Evalenv.lookupRelationGlobalEnv env (Ast.Var rel2) with
  | Some d1, Some d2 ->
    let length1 = MapUtils.int_map_size d1 in
      if debug_level 3
        then Printf.printf "Builtins.unionRelations: joining relations.\n"
        else ();
      appendRelations length1 d2 d1
  | Some _, None -> raise (Unbound (rel2, "Builtins.unionRelations: rel2"))
  | None, Some _ -> raise (Unbound (rel1, "Builtins.unionRelations: rel1"))
  | None, None ->
      raise (Unbound ("relation in union", "Builtins.unionRelations"))
                          *)

(*
(* Does join of 2 relations. *)
let joinRelations data_name env rname1 rname2 byCol =
  let joinRelations (rel1 : Ast.datafields list) (rel2 : Ast.datafields list) =
    (* Joins a record to a list of records. *)
    let join (r1 : Ast.datafields) (rlist : Ast.datafields list) =
      let r1_records =
        List.remove_assoc byCol r1 in
      let combine_records accum record =
      let new_record = List.append r1_records record in
        new_record::accum in
      List.fold_left combine_records [] rlist in
    (* Joins a record with things that match over a list of records. *)
    let joinRecord joinTo record_accum cur_record : Ast.datafields list =
      let colEntry = AstUtils.get_field cur_record byCol in
      let eqCol (record : Ast.datafields) =
        AstEq.eq_exps colEntry
                      (AstUtils.get_field record byCol) in
      let rel2_eq_records = List.filter eqCol rel2 in
        List.append (join cur_record rel2_eq_records) record_accum in
    List.fold_left (joinRecord rel2) [] rel1 in
  match Evalenv.lookupDataGlobalEnv env rname1,
        Evalenv.lookupDataGlobalEnv env rname2 with
  | Some d1, Some d2 ->
      let new_relation =
        (* Have to update the order ID's on the second guy... *)
        { Ast.recordlist = joinRelations d1.Ast.recordlist d2.Ast.recordlist } in
        processConstraints env data_name new_relation.Ast.recordlist []
  | Some _, None -> raise (Unbound rname2)
  | None, Some _ -> raise (Unbound rname1)
  | None, None -> raise (Unbound "relation in join") *)

  end;;
