(*
  File          : evalExp.ml
  Last updated  : 10.7.09
  Maintainer    : jeanyang@csail.mit.edu

  This file contains the function for expression evaluation.
*)
open AstUtils
open Error
open Evalenv

module Ast = Ast.Ast

(*
let rec evalSelect
      (env : Evalenv.evalenv)
      (cenv : Evalenv.constraintenv ref)
      (t : Ast.typerep)
      (args : Ast.exp list)
      : Ast.taggedExp =
  if List.length args == 4
   then
    (* First argument is field name. *)
     let (fieldname : string) =
       match List.nth args 0 with
         | Ast.VarExp (Ast.Var f) -> f
         | _ ->
             raise
               (Unexpected ("Evalenv.processConstraint: " ^
                            "expect field for first arg.")) in
    (* Second argument is the relation. *)
     let (arg1, _sym) = evalExp env cenv (List.nth args 1) t in
     let (relation : Ast.relation ref) =
       match arg1 with
          | Ast.PrimExp (Ast.RelationExp (n, ty, rel)) -> rel
          | _ ->
            raise
              (Unexpected ("EvalExp.evalSelect: Second argument should be " ^
                           "a relation.")) in
    (* Third argument is the column name. *)
     let (colname : string) =
       match List.nth args 2 with
         | Ast.VarExp (Ast.Var f) -> f
         | _ ->
             raise
               (Unexpected ("Evalenv.processConstraint: " ^
                            "expect field for first arg.")) in
    (* Fourth argument is the value it's supposed to be equal to. *)
     let (arg3, _sym) =
       evalExp env cenv (List.nth args 3) Ast.IntType in
     let (v : int option) =
       match arg3 with
         | Ast.PrimExp (Ast.Integer i) -> Some i
         | Ast.Symbolic _ -> None
         | _ ->
             raise
               (Unexpected ("Evalenv.processConstraint: " ^
                            "expect int for first arg.")) in
      let (exps, sym) =
        match v with
         | Some value ->
             (try
                let find_record _uid cur_record (accum, sym) =
                  (try
                     (* TODO: This doesn't account for symbolic equality. *)
                     if AstEq.eq_exps
                          (fst (StringMap.find colname !cur_record))
                          (Ast.PrimExp (Ast.Integer value))
                     then
                       let (exp, cur_sym) =
                         StringMap.find fieldname !cur_record in
                        (exp::accum, cur_sym || sym)
                    else (accum, sym)
                   with Not_found -> (accum, sym)) in
                  List.fold_left find_record !relation ([], false)
              with Not_found -> ([], false))
         | None -> ([], false) in
                  (Ast.PrimExp (Ast.ListExp (t, exps)), sym)
  else raise (Unexpected "EvalExp.evalSelect: Expected 4 arguments.")
 *)

(* Evaluates a binary arithmetic expression. *)
let evalArithop
      (op : Ast.binop) (e's : Ast.exp ref list)
      (* TODO: write types *)
      (evalFun)
      : Ast.taggedExp ref =
  let commutative op : bool =
    match op with
      | Ast.Plus | Ast.Times -> true
      | Ast.Minus | Ast.Divide -> false in
  let opr =
    match op with
      | Ast.Plus -> (+)
      | Ast.Minus -> (-)
      | Ast.Times -> ( * )
      | Ast.Divide -> (/) in
  (* Get the integer part of the expression. *)
  let get_int e =
    match !(Ast.get_exp e) with
      | Ast.PrimExp (Ast.Integer i) -> i
      | _ -> raise (Unexpected "non-int binop") in
  let (flattened : Ast.taggedExp ref list ref) = ref [] in
  (* Add everything in backwards order. *)
  let flatten (e : Ast.taggedExp ref) : unit =
    (match !(Ast.get_exp e) with
       | Ast.Arithop (bop, e's) ->
           if bop = op
           then flattened :=
                List.rev_append (List.map (fun x -> ref x) e's) !flattened
           else flattened := e::!flattened
       | _ -> flattened := e::!flattened);
    if commutative op
    then ()
    else flattened := List.rev !flattened in
  (* Get resulting expressions. *)
  let result_e's =
    (* TODO: fix type *)
    List.map evalFun
    (* (fun (e, _sym) -> evalExp env cenv e Ast.UnknownType conc) *) e's in
    (* Flatten the expression. *)
    List.iter flatten (List.rev result_e's);
    if commutative op
    then
      let (symbolic, concrete) =
        List.partition Ast.get_sym !flattened in
        (match symbolic with
           (* If everything is concrete, then concretize. *)
           | [] ->
               let concrete_ints = List.map get_int concrete in
               let result =
                 List.fold_left opr
                   (List.hd concrete_ints) (List.tl concrete_ints) in
                 ref (ref (Ast.PrimExp (Ast.Integer result)), false)
           (* Otherwise concretize as much as possible. *)
           | _ -> 
               let result =
                 (match concrete with
                    | [] -> symbolic
                    | _ ->
                        let concrete_ints = List.map get_int concrete in
                        let result_int =
                          List.fold_left opr 
                            (List.hd concrete_ints) (List.tl concrete_ints) in
                        let result_exp =
                          ( ref (Ast.PrimExp (Ast.Integer result_int))
                            , false) in
                          ((ref result_exp)::symbolic)) in
                 ref
                   (ref (Ast.Arithop (op, (List.map (fun x -> !x) result)))
                    , true))
        else
          let sym = List.exists Ast.get_sym result_e's in
            ref
              ( ref (Ast.Arithop (op, (List.map (fun x -> !x) result_e's)))
                , sym)
(* Evaluates a boolean expression. *)
let evalBoolop (op : Ast.boolop) (e's : Ast.taggedExp list) (evalFun) (* TODO *)
      : Ast.taggedExp ref =
  let commutative op : bool =
    match op with
      | Ast.Equiv | Ast.Nequiv | Ast.Or | Ast.And -> true
      | Ast.Lt | Ast.Gt -> false in
  (* Whether the operator expects boolean arguments. *)
  let bool_args op : bool =
    match op with
      | Ast.Or | Ast.And -> true
      | Ast.Equiv | Ast.Nequiv | Ast.Lt | Ast.Gt -> false in
  let opr_bool () =
    match op with
      | Ast.Or -> (||)
      | Ast.And -> (&&)
      | _ ->
          raise
            (InternalError
               "opr_bool: Expected operator that takes boolean arguments.") in
  let opr_int () =
    match op with
      | Ast.Equiv -> (==)
      | Ast.Nequiv -> (!=)
      | Ast.Lt -> (<)
      | Ast.Gt -> (>)
      | _ ->
          raise
            (InternalError
               "opr_int: Expected operator that takes int arguments.") in
  (* Get the integer part of the expression. *)
  let eval_opr (accum_exp : Ast.primexp ref) (e : Ast.primexp ref)
        : Ast.primexp ref =
    if bool_args op
    then
      (match !accum_exp, !e with
         | Ast.Bool ab, Ast.Bool b -> ref (Ast.Bool ((opr_bool ()) ab b))
         | _ -> raise (Unexpected "non-bool"))
    else
      (match !accum_exp, !e with
         | Ast.Integer ai, Ast.Integer i -> ref (Ast.Bool ((opr_int ()) ai i))
         | _ -> raise (Unexpected "non-int binop")) in
  let (flattened : Ast.taggedExp ref list ref) = ref [] in
  (* Add everything in backwards order. *)
  let flatten (e : Ast.taggedExp ref) : unit =
    match !(Ast.get_exp e) with
      | Ast.Boolop (bop, e's) ->
          if bop = op
          then flattened :=
               List.rev_append (List.map (fun x -> ref x) e's) !flattened
          else flattened := e::!flattened
      | _ -> flattened := e::!flattened in
  (* Get the primitive expression out of the AST expression. *)
  let get_primexp e =
    match !(Ast.get_exp e) with
      | Ast.PrimExp pe -> ref pe
      | _ -> raise (Unexpected "non-primexp") in
  (* Get resulting expressions. *)
  let result_e's =
    (* TODO: fix type *)
    List.map evalFun
    (* (fun (e, _sym) -> evalExp env cenv e Ast.UnknownType conc)*) e's in
    (* Flatten the expression. *)
    List.iter flatten (List.rev result_e's);
    if false (* commutative op *)
    then
      let (symbolic, concrete) =
        List.partition Ast.get_sym !flattened in
      let concrete_prims = List.map get_primexp concrete in
        (match symbolic with
           (* If everything is concrete, then concretize. *)
           | [] ->
               let result : Ast.primexp ref =
                 List.fold_left eval_opr
                   (List.hd concrete_prims) (List.tl concrete_prims) in
                 ref (ref (Ast.PrimExp !result), false)
           (* Otherwise concretize as much as possible. *)
           | _ ->
               let result =
                 (match concrete with
                    | [] -> symbolic
                    | _ ->
                        let result =
                          List.fold_left eval_opr
                            (List.hd concrete_prims)
                            (List.tl concrete_prims) in
                        let result_exp =
                          ( ref (Ast.PrimExp !result)
                            , false) in
                          ((ref result_exp)::symbolic)) in
                 ref
                   (ref (Ast.Boolop (op, (List.map (fun x -> !x) result)))
                    , true))
        else
          let sym = List.exists Ast.get_sym result_e's in
            ref
              ( ref (Ast.Boolop (op, (List.map (fun x -> !x) result_e's)))
                , sym )
