(**
    Module        : Evalenv
    Last updated  : 2.8.10
    Maintainer    : jeanyang@csail.mit.edu
  *)
open Ast
open AstUtils

module VarEnv : Hashtbl.S with type key = Ast.varname
module TypeEnv : Hashtbl.S with type key = Ast.typename

(** Evaluation environment. *)
module Evalenv :
  sig
    (** For storing information about data values/user-defined types along with
        their associated constraints. *)
    type tyenv = Ast.typerep VarEnv.t
    and evalenv = {(* Map type names to type constraints. *)
                     typeconstraints    : ((Ast.varname * Ast.cdecl list) VarEnv.t)
                                            ref
                    (* Map variable names to types. *)
                   ; tyenv              : (Ast.typerep VarEnv.t) ref
                    (* Map type names to type information *)
                   ; tydefenv           : (Ast.userty VarEnv.t) ref
                    (* Stores information about function calls. *)
                   ; functions          : ((Ast.fundecl * evalenv ref)
                                             VarEnv.t) ref }
    
    (** Default empty environment. *)
    val empty_evalenv : evalenv

    (** Type names -> type constraints. *)
    (** Function declarations. *)
    val addFunDecl : evalenv ref -> Ast.varname -> Ast.fundecl -> evalenv ref
                  -> unit
    val lookupFunDecl : evalenv ref -> Ast.varname ->
      (Ast.fundecl * evalenv ref) option

    (** Var names -> type representations. *)
    val bindTypeEnv : evalenv ref -> Ast.varname -> Ast.typerep -> unit
    val lookupType : evalenv ref -> Ast.var -> Ast.typerep option
    
    (** Type names -> constraints. *)
    val addTypeDecl : evalenv ref -> Ast.userty -> Ast.varname * Ast.cdecl list
                    -> unit
    val lookupTypeDeclByData : evalenv ref -> Ast.varname -> Ast.userty option
    val lookupNumFieldsByData : evalenv ref -> Ast.varname -> int

    (** Derives type-level constraints for variable name passed as arg. *)
    (* val addTypeConstraints : evalenv ref -> constraintenv ref -> Ast.varname
                          -> Ast.taggedExp ref -> unit *)

    (** Var names -> expressions *)
    val bindValueEnv : evalenv ref -> Ast.var -> unit
    val lookupValueEnv : evalenv ref -> Ast.var -> unit

    (** Var names -> constraints. *)
    val addConstraintEnv : Ast.exp ref -> unit
    val addConstraintsEnv : Ast.exp ref list -> unit
  end
