'''
Processing files that have record struction
  [household] [line] [age] [married] [spouse]
'''
import sys
import time

input_rel = []
input_rel_by_household = {}
ref_rel_by_household = {}

def readFromFile(refname, filename):
  f = open(refname, 'r')
  data_lines = f.readlines()
  for line in data_lines:
    line = line.rstrip();
    line_fields = line.split()
    if len(line_fields) == 6:
      # Also add to our dictionary by household.
      household = line_fields[0]
      if ref_rel_by_household.has_key(household):
        ref_rel_by_household[household].append(line_fields)
      else:
        ref_rel_by_household[household] = [line_fields]

  f = open(filename, 'r')
  data_lines = f.readlines()
  for line in data_lines:
    line = line.rstrip();
    line_fields = line.split()
    if len(line_fields) == 6:
      # Add to our list of input records.
      input_rel.append(line_fields)
      # Also add to our dictionary by household.
      household = line_fields[0]
      if input_rel_by_household.has_key(household):
        input_rel_by_household[household].append(line_fields)
      else:
        input_rel_by_household[household] = [line_fields]

  print time.clock()

'''
Computes the sum age of people who are married who have someone in their
household who gets free lunch.
'''
def sumInputRelCondAge():
  age_sum = 0
  for entry in input_rel:
    married = False

    # Figure out if the current person is married.
    if entry[3] == "1":
      married = True
    elif entry[3] == "!!":
      # If the entry is missing, loop over the list of records in the same
      # household to find if someone lists the current person as their spouse.
      for housemate in input_rel_by_household[entry[0]]:
        married = housemate[4] == entry[1]

    # Figure out if the current person has anybody in the household getting hot
    # lunch.
    if not married:
      gets_hot_lunch = False
      if entry[5] == "!!":
        for housemate in input_rel_by_household[entry[0]]:
          housemate_hotlunch = housemate[5]
          if housemate_hotlunch != "!!":
            gets_hot_lunch = int(housemate_hotlunch) > 0
      else:
        gets_hot_lunch = int(entry[5]) > 0

    if married and gets_hot_lunch:
      # Now the age might be missing too.
      if entry[2] == "!!":
        age = 0
        for housemate in ref_rel_by_household:
          if housemate[2] != "!!":
            age = int(housemate[2])
      else:
        age = int(entry[2])
      age_sum = age_sum + age
  print age_sum

def main():
  # Read from file.
  readFromFile(sys.argv[1], sys.argv[2])
  sumInputRelCondAge()
  print time.clock()

if __name__ == "__main__":
  main()
