(* processYices.ml *)
open Error
open MapUtils

(* Mapping back to environment *)

(* Printing *)
let printOutput (yices_out : Yices.yexp list) (results : Ast.relation) =
  (* helper function *)
  let get_oid record = match List.assoc "oid" record with
    | Ast.PrimExp (Ast.Integer i) -> i
    | _ -> raise (Unexpected "Non-integer OID") in
  (* Look for whether the variable corresponds to a result. *)
  let has_result str yexp = match yexp with
    | Yices.YEqual (var, _value) ->
        try Str.search_forward (Str.regexp str) var 0; true
        with Not_found -> false in
  (* Procedure for serializing results. *)
  let pickle_results (count : int) (result : Ast.datafields) : string =
    let pickle_field field_name exp accum =
      let new_str = match exp with
      | Ast.PrimExp (Ast.Integer i) ->
          (String.concat "\t" [field_name; string_of_int i])
      | _ ->
          let varname = Utils.to_record_name ("result", field_name, count) in
          try match List.find (has_result varname) yices_out with
              | Yices.YEqual (_var, i) ->
                  (String.concat "\t" [field_name; string_of_int i])
          with Not_found -> String.concat "\t" [field_name; "??"] in
        String.concat "" [new_str; accum] in
    let pickled = StringMap.fold pickle_field result "" in
      String.concat "" ["-----\n"; pickled] in

  let concat strmap =
    let accum = ref [] in
    let inc_accum _ a = accum := a::!accum in
      IntMap.iter inc_accum strmap;
      !accum in
    print_endline "Printing results";
    List.map print_endline (concat (IntMap.mapi pickle_results results));
    ()
