(* 
    Module      : server.ml
    Maintainer  : jeanyang@csail.mit.edu
    Updated     : 1.11.11

    This contains code for the social network server.
*)

open AstUtils
open Error

let split = Str.split (Str.regexp_string " ")

(* Process client requests. *)
let process_client_request (req : string) : unit =
  let req_elts = split req in
    if List.length req_elts > 0
    then
      match List.hd req_elts with
        | "search" -> print_endline "search"
        | _ -> print_endline "other"
    else print_endline "Bad request."

let sock_send sock str =
            let len = String.length str in
    Unix.send sock str 0 len []

(* Create a unix domain socket server. *)
let sock_name = (Unix.gethostbyname (Unix.gethostname())).Unix.h_addr_list.(0)
let server = Unix.socket Unix.PF_INET Unix.SOCK_STREAM 0
let buffer_size = 256
let buffer = String.make buffer_size '\000'

let run_server () =
    (try Unix.bind server (Unix.ADDR_INET (sock_name, 1029))
    with err ->
      (match err with
         | Unix.Unix_error (e, fn, param) ->
             print_endline (Unix.error_message e);
             raise err
         | _ -> raise err));
    Unix.listen server 10;
    while true do
      (try
        (* Client is a file_descr; sock_addr is the address of the connecting
         * client. *)
        let (client, addr) = Unix.accept server in
        (* Read data. *)
        let chars_read =
          (try Some (Unix.read client buffer 0 buffer_size)
           with Unix.Unix_error (error, _, _) ->
              prerr_endline (Unix.error_message error);
              None) in
          (* Examine characters we read. *)
          (match chars_read with
            | None | Some 0 ->
              (* This would be the end of file, so close the client *)
                print_endline "read no chars";
                Unix.close client
            | Some chars_read ->
                print_endline buffer;
                (* Process the buffer. *)
                process_client_request buffer;
                (* Clear the buffer. *)
                String.fill buffer 0 buffer_size '\000';
                (* Send back the result. *)
                sock_send client "received";
                ())
       with err ->
         match err with
           | Unix.Unix_error (err, fn, param) ->
               print_endline (Unix.error_message err)
           | _ -> print_endline "other error")
    done

let main () =
  run_server ()

let _ = Printexc.print main ()

