open AstBackend
open AstUtils
open BackendInterface
open Error
open TestUtils

module Ast = Ast.Ast
module Evalenv = Evalenv.Evalenv
module Evaluator = Evaluator.Evaluator

let result = TestUtils.gen_var ()
let mkIntAssign exp =
  [ Ast.AssignDecl (result, Ast.IntType, (Ast.RhsExp exp)) ]

(* TEST PROGRAM 0 *)
(* Constructs a simple program that returns a single integer value. *)
let result0 = raise Undefined
                (*
  fun _ -> AstBackend.mkAstExpNode (ref (Ast.PrimExp (Ast.Integer 42))) *)
let program0 =
  fun _ -> mkIntAssign (Ast.PrimExp (Ast.Integer 42))

(* TEST PROGRAM 1 *)
let result1 = raise Undefined
                (*
  fun _ -> AstBackend.mkAstExpNode (ref (Ast.PrimExp (Ast.Integer 84))) *)
let program1 =
  fun _ ->
    let sum_exp = TestUtils.mkArithop Ast.Plus
                    (ref (Ast.PrimExp (Ast.Integer 42)))
                    (ref (Ast.PrimExp (Ast.Integer 42))) in
      mkIntAssign !sum_exp

(* TEST PROGRAM 2: test_sym_var *)
let test = TestUtils.gen_var ()
let result2 = raise Undefined
                (*
  fun _ -> AstBackend.mkAstExpNode (ref (Ast.Symbolic test)) *)
let program2 = fun _ -> mkIntAssign (Ast.Symbolic (test, Ast.ref_true))

(* TEST PROGRAM 3: test_sym_exp *)
let result3_exp =
  TestUtils.mkArithop Ast.Plus
    (ref (Ast.Symbolic (test, Ast.ref_true)))
    (ref (Ast.PrimExp (Ast.Integer 42)))
let result3 = raise Undefined (* fun _ -> AstBackend.mkAstExpNode result3_exp *)
let program3 = fun _ -> mkIntAssign !result3_exp

(* TEST PROGRAM 4 *)
(* An empty relation. *)
let testty = TestUtils.gen_var ()
(*
let rel4 = BackendInterface.mkRelation result 1
let result4 = Ast.ExpId rel4
let type4 = Ast.RelationType testty
let program4 =
  [ Ast.AssignDecl (result, type4, (Ast.RhsExp result4)) ]
                  *)

(* TEST PROGRAM 5 *)
(* A non-empty relation. *)
let rel_list_5 =
    [ (0, ["1"])
    ; (1, ["3"])
    ; (2, ["5"])
    ; (3, ["7"]) ]
let result5_name = TestUtils.gen_var ()
let rel =
  fun _ -> TestUtils.mkRelFromList result5_name testty 1 rel_list_5
let result5 = raise Undefined
                (*
  fun _ -> BackendInterface.lookupVal result5_name *)
let type5 = Ast.RelationType testty
let program5 =
  fun _ ->
    rel ();
    [ Ast.AssignDecl
      (result, type5, (Ast.RhsExp (Ast.VarExp (Ast.Var result5_name)))) ]

(**************)
(*** CHECKING *)
(**************)
(* Check that the result expression is the same as the intended result
 * expression. *)
let checkResultExp (env : Evalenv.evalenv ref) : unit =
  print_endline "";
  print_endline "CHECKING RESULT...";
  Evalenv.lookupValueEnv env (Ast.Var result);
  if BackendInterface.eqExps ()
    then print_endline "Test passed."
    else raise (TestUtils.TestFail "exps not equal") 
           (*
  | Some te ->
      print_endline "\tMake sure expressions equal:";
      print_endline ("\t" ^ (AstPrint.pickle_exp te));
      print_endline ("\t" ^ (AstPrint.pickle_exp (ref r)));
      TestUtils.assert_exp_eq te (ref r)
  | None -> raise (Impossible "result unbound") *)

let checkResultType (env : Evalenv.evalenv ref) (ty : Ast.typerep) : unit =
    match Evalenv.lookupType env (Ast.Var result) with
  | Some t -> TestUtils.assert_ty_eq t ty
  | None -> raise (Unbound ("result", "test/simpleResult.checkResultType"))

let run_simple_tests () =
  let tests = [ ("result_int", program0, result0, Ast.IntType)
              ; ("result_int_sum", program1, result1, Ast.IntType)
              ; ("result_sym_var", program2, result2, Ast.IntType)
              ; ("result_sym_exp", program3, result3, Ast.IntType)
(*              ; ("result_empty_relation", program4, result4, type4) *)
              ; ("result_relation", program5, result5, type5) ] in
  let run_test (test_name, test_program, test_result, test_ty) =
    print_endline ("***\n" ^ test_name ^ " running...");
    let denv = ref Evalenv.empty_evalenv in
    Evaluator.evaluate denv (test_program ());
    print_endline "Evaluating result...";
    test_result ();
    checkResultExp denv;
    checkResultType denv test_ty;
(*     BackendInterface.clearEnv (); *)
    print_endline (test_name ^ " passed...\n") in
  print_endline "*** RUNNING SIMPLE TESTS ***";
  List.iter run_test tests
