open AstUtils
open Builtins
open Error
open Evalenv
open SortUtils
open TestUtils

module Ast = Ast.Ast

type inputdata = (int * (string * string) list) list

let inc_field = "timestamp"

(*
  data indata { timestamp :: ~int }
  @@ indata {
    missing inc_timestamp.
  }
*)
let construct_initial_env () : Evalenv.evalenv =
  raise Undefined
    (*
  let env = ref Evalenv.empty_evalenv in
  let indata_decl =
        Ast.UserType { Ast.dataname = "indata"
                     ; Ast.fieldinfo = [ (inc_field, true, Ast.IntType) ] } in
  (* This is associated with the hack constraint processor. *)
  let indata_constraint =
        Ast.ConstraintExp ("inc_timestamp", inc_field, Constraint.CTrue) in
    env := Evalenv.addTypeDecl !env indata_decl;
    env := Evalenv.addConstraint !env "indata" [indata_constraint];
    (* env := Evalenv.bindValueGlobalEnv !env (Ast.Var "cur_data") input_relation; *)
    !env *)

let get_sorted_result env input0 input1 output_name =
  (* Take union of inputs. *)
  let crel0 = TestUtils.list2relation "rel0" input0 in
  let crel1 = TestUtils.list2relation "rel1" input1 in
  let crel0_exp = Ast.PrimExp (Ast.RelationExp ("rel0", crel0)) in
  let crel1_exp = Ast.PrimExp (Ast.RelationExp ("rel1", crel1)) in
  let rel_env =
      Evalenv.bindValueGlobalEnv
        (Evalenv.bindValueGlobalEnv env (Ast.Var "rel0") crel0_exp)
        (Ast.Var "rel1") crel1_exp in
  let new_rel = Builtins.unionRelations rel_env "rel0" "rel1" in
  let new_env =
    Evalenv.bindValueGlobalEnv rel_env (Ast.Var "indata")
      (Ast.PrimExp (Ast.RelationExp ("indata", new_rel))) in
  let result_relation =
        SortUtils.sortRecords new_env output_name inc_field new_rel in
  let sorted_exp = Ast.PrimExp (Ast.RelationExp ("result", result_relation)) in
  let env' = Evalenv.bindValueGlobalEnv env (Ast.Var output_name) sorted_exp in
  let (result_env, result_exp) = Oracle.Oracle.concretize env' sorted_exp in
    result_exp

(* Test input with all concrete records. *)
let test_concrete_0 () =
  let concrete_data_0 =
    [ (0, [(inc_field, "1")])
    ; (1, [(inc_field, "3")])
    ; (2, [(inc_field, "5")])
    ; (3, [(inc_field, "7")]) ] in
  let concrete_data_1 =
    [ (0, [(inc_field, "2")])
    ; (1, [(inc_field, "4")])
    ; (2, [(inc_field, "6")])] in

  let init_env = construct_initial_env () in
  let result_exp =
        get_sorted_result init_env concrete_data_0 concrete_data_1 "cur_data" in

   let check_field relation uid nid =
    match RelationUtils.get_index_field "nid" uid relation with
  | Some e -> (match e with
      | Ast.PrimExp (Ast.Integer i) ->
          if i == nid
            then ()
            else (Printf.printf "ID: %d;\t NID: %d;\t EXPECTED: %d\n" uid i nid;
                  Printf.printf "%s\n" (AstPrint.pickle_exp result_exp);
                  raise (TestUtils.TestFail "fail"))
      | _ -> raise (TestUtils.TestUnexpected "bad"))
  | None -> raise (TestUtils.TestFail "bad") in
    match result_exp with
  | Ast.PrimExp (Ast.RelationExp (_n, relation)) ->
    check_field relation 0 4;
    check_field relation 1 5;
    check_field relation 2 6;
    check_field relation 3 7;
    check_field relation 4 1;
    check_field relation 5 2;
    check_field relation 6 3
  | _ -> raise (TestUtils.TestUnexpected "result was not a relation")

(* Test input with all concrete records. *)
let test_symbolic_0 () =
   let concrete_data_0 =
    [ (0, [(inc_field, "1")])
    ; (1, [(inc_field, "3")])
    ; (2, [(inc_field, "??")])
    ; (3, [(inc_field, "7")]) ] in
  let concrete_data_1 =
    [ (0, [(inc_field, "2")])
    ; (1, [(inc_field, "4")])
    ; (2, [(inc_field, "6")])] in

  let init_env = construct_initial_env () in
  let result_exp =
        get_sorted_result init_env concrete_data_0 concrete_data_1 "cur_data" in

   let check_field relation uid nid =
    match RelationUtils.get_index_field "nid" uid relation with
  | Some e -> (match e with
      | Ast.PrimExp (Ast.Integer i) ->
          if i == nid
            then ()
            else (Printf.printf "ID: %d;\t NID: %d;\t EXPECTED: %d\n" uid i nid;
                  Printf.printf "%s\n" (AstPrint.pickle_exp result_exp);
                  raise (TestUtils.TestFail "fail"))
      | Ast.PrimExp (Ast.RelationExp (_n, relation)) ->
          Printf.printf "%s\n" (AstPrint.pickle_exp result_exp)
      | _ -> raise (TestUtils.TestUnexpected "bad"))
  | None -> raise (TestUtils.TestFail "bad") in
    match result_exp with
  | _ -> raise (TestUtils.TestUnexpected "result was not a relation")

let test_sort () = ()
                     (*
  test_concrete_0 ();
  test_symbolic_0 (); *)
