(**
    File         : testUtils.ml
    Last updated : 12.14.09

    Testing infrastructure.
*)
open AstBackend
open AstUtils
open BackendInterface
open Error

module Ast = Ast.Ast
module Evalenv = Evalenv.Evalenv

module TestUtils =
  struct
    (***************)
    (* Exceptions. *)
    (***************)
    exception TestUnexpected of string
    exception TestFail of string

    (**************)
    (* Variables. *)
    (**************)
    let var_count = ref 0
    let gen_var () =
      let new_var = !var_count in
        var_count := !var_count + 1;
        new_var
    let result_name = gen_var ()

    let mkConcreteResult
          (name : Ast.varname) (ty : Ast.typerep) (v : Ast.varname) =
      Ast.AssignDecl (name, ty, Ast.Concretize (0, (Ast.VarExp (Ast.Var v))))

    (* Equality testing. *)
    let assert_exp_eq e1 e2 =
        if AstEq.eq_exps e1 e2
          then ()
          else
            (print_endline "The following expressions were not equal:";
            print_endline (AstPrint.pickle_exp e1);
            print_endline (AstPrint.pickle_exp e2);
            raise (TestFail "assertion not equal"))

      let assert_ty_eq e1 e2 =
        if AstEq.eq_types e1 e2
          then ()
          else raise (TestFail "assertion not equal")

    (*********************)
    (* AST construction. *)
    (*********************)
    let mkFieldIndex (var : Ast.varname) (field : Ast.fieldname) =
      ref (Ast.VarExp (Ast.FieldIndexVar (var, field)))
    let mkArithop op elt1 elt2 = ref (Ast.Arithop (op, elt1, elt2))
    let mkBoolop op elt1 elt2 = ref (Ast.Boolop (op, elt1, elt2))

    (* Makes a relation. *)
    let mkRelation
          (n : Ast.varname) (ty : Ast.typename) (numfields : int)
          (rel : (int * (Ast.exp ref list)) list) : unit =
      raise Undefined
        (*
      let make_row (id, exps) : unit =
        let field = ref 0 in
        let make_elt exp =
          AstBackend.mkAstExpNode exp;
          BackendInterface.bindRelationVal n id !field;
            field := !field + 1 in
          List.iter make_elt exps in
        BackendInterface.mkRelation n numfields;
        List.iter make_row rel *)
    
    let mkRelFromList
          (n : Ast.varname) (ty : Ast.typename) (numfields : int)
          (exps : (int * string list) list) =
      raise Undefined
        (*
      debug_print 2 ("TestUtils.mkRelFromList: " ^ string_of_int n);
        let make_row (id, strs) : unit =
          let field = ref 0 in
          let make_elt (str : string) =
            let exp =
              try (BackendInterface.Int (int_of_string str))
              with _ -> (BackendInterface.Sym (gen_var ())) in
            BackendInterface.createAtomEid exp;
            BackendInterface.bindRelationVal n id !field;
              field := !field + 1 in
            List.iter make_elt strs in
          (* let rel = ref (Ast.IntMap.create (List.length exps)) in *)
        BackendInterface.mkRelation n numfields;
        List.iter make_row exps *)

    (* Constraint construction. *)
(*    let mkACFieldIndex (var : Ast.varname) (field : Ast.fieldname) =
      ref (Ast.ACVarExp (Ast.FieldIndexVar (var, field))) *)
    let mkACBoolop op elt1 elt2 =
      ref (Ast.Boolop (op, elt1, elt2))
  end
