(**
    Module       : Testbasic
    Last updated : 12.16.09

    Tests the basic functionalities (environments, etc.) of the LogLog
    interpreter.
*)
open AstUtils
open BackendInterface
open Error
open Evalenv

exception TestFail of string

module Ast = Ast.Ast
module TestUtils = TestUtils.TestUtils

(* Tests the environment. *)
let test_env () =
  (* Variables. *)
  let v = TestUtils.gen_var () in
  let test0 = TestUtils.gen_var () in
  let test1 = TestUtils.gen_var () in

  let dummy_int = Ast.PrimExp (Ast.Integer 42) in
  let dummy_sym = Ast.Symbolic (v, Ast.ref_true) in
  (* Values to try adding to our environment. *)
  let test_values = 
    [ (ref Evalenv.empty_evalenv, Ast.Var test0, dummy_int)
    ; (ref Evalenv.empty_evalenv, Ast.Var test1, dummy_sym) ] in
    (*
    ; (env_w_record, Ast.RecordIndexVar (0, tr, tf), dummy_int)
    ; (env_w_record, Ast.RecordIndexVar (0, tr, tf), dummy_sym)] in *)

  (*
   * Tests that the given variable is mapped to the given value.
   *)
  let test_value_env (new_env, test_var, test_value) =
    print_endline "";
    print_endline "RUNNING BASIC TEST...";
    EvalExp.evalExp new_env (ref test_value) Ast.UnknownType;
    BackendInterface.dup ();
    Evalenv.bindValueEnv new_env test_var;
    Evalenv.lookupValueEnv new_env test_var;
    if BackendInterface.eqExps ()
      then (print_endline "Values matched:";
            print_endline (AstPrint.pickle_exp (ref test_value)))
      else (print_endline
              ("Values did not match for " ^ AstPrint.pickle_var test_var)) in
  let test_local_env () =
    let test_gvalue = Ast.PrimExp (Ast.Integer 0) in
    let test_lvalue = Ast.PrimExp (Ast.Integer 1) in
    let env_w_gvalue =
      let new_env = ref Evalenv.empty_evalenv in
        EvalExp.evalExp new_env (ref test_gvalue) Ast.UnknownType;
        BackendInterface.dup ();
        Evalenv.bindValueEnv new_env (Ast.Var test0);
        new_env in
    let env_w_lvalue =
      let new_env = ref Evalenv.empty_evalenv in
        EvalExp.evalExp new_env (ref test_lvalue) Ast.UnknownType;
        BackendInterface.dup ();
        Evalenv.bindValueEnv new_env (Ast.Var test1);
        new_env in
      print_endline "Looking up lvalue...";
      Evalenv.lookupValueEnv env_w_lvalue (Ast.Var test1);
      if BackendInterface.eqExps ()
        then ()
        else raise (TestFail "value env");
      print_endline "Looking up gvalue...";
      Evalenv.lookupValueEnv env_w_gvalue (Ast.Var test0);
      if BackendInterface.eqExps ()
        then ()
        else raise (TestFail "value env") in
  List.iter test_value_env test_values;
  print_endline "Finished value environment tests...";
  print_endline "Starting local environment tests...";
  test_local_env ();
  ()

let main () =
  set_debug 3;
  test_env ()

let _ = Printexc.print main ()
