(*
  View whole thing as declarative language where inputs are clear, so the order
  of the whole thing is also clear.

  It seems like we can just realize that, for instance, if timestamp is missing,
  we can just look at constraints and figure out what we would need to carry
  through the program to make things work.  Whenever we have conditionals, we
  would need to be able to handle the symbolic values as well as the actual
  values.

  Questions:
  - What would static constraint solving do for us?
  - Why would we want to look at whether predicates are tautologies ahead of
    time?

  I am assuming a ton of syntactic sugar here, but it should be good enough to
  look at things for now.
*)
(* Declare data to be ordered list of inputs. *)
data cvsdata { fid :: int
             , uid :: int
             , timestamp  :: ~int
             , rn  :: int }
input input_data :: cvsdata list

(* These assertions would generate constraints based on the positions of
   elements in the input list.
*)
@@ input_data {
  increasing uid.
  increasing timestamp.
}

(* Sort data based on timestamps.  We should have the sort generate implicit
   constraints based on what it's sorting by. *)
(* let sorted_cvsdata :: cvsdata list = sort input_data timestamp *)

(* 2 requirements on a change set: they occur within some timeframe and were
   made by the same user. *)
type changeset = cvsdata list

(* Assertions on things that have properties of that type. *)
@@ changeset {
  unique fid.
  same uid.
  range ts 5.
}

(* Code for making change sets. *)
(* Want notion of accumlation built into the language *)
(* make_changesets :: [cvsdata] -> [changeset] *)
function make_changesets (cdata :: cvsdata list) = {
  let groupid = 0;

  let curuid = 100;
  let curtimestamp = 0;

  foreach cvsentry in cdata {
    (* If the timestamps are far enough apart, start a new guy. *)
    if ((cvsentry.timestamp - curtimestamp) > 5) {
      let groupid = groupid + 1;
      .let cvsentry.gid = groupid;
    } else {
      if (cvsentry.uid != curuid) {
        let groupid = groupid + 1;
        .let cvsentry.gid = groupid;
      } else {
        let groupid = groupid;
      }
    }
    let curtimestamp = cvsentry.timestamp;
  }
}
let change_sets :: changeset list = make_changesets input_data (* sorted_cvsdata *)
