(**
    File          : varenv.ml
    Maintainer    : jeanyang@csail.mit.edu
    Last updated  : 11.11.10

    Variable environment--maps the integers variables back to the strings for
    printing and other purposes.
*)

open Error
module Ast = Ast.Ast

module StringTbl = Hashtbl.Make
                  (struct type t = string
                     let equal s1 s2 = String.compare s1 s2 == 0
                     let hash v = 0
                   end)
module StringMap = Map.Make (String)
type varenv = { fieldenv : (int StringMap.t) StringTbl.t
              ; varnameenv : int StringTbl.t
              ; typenameenv : int StringTbl.t }

(* Initialize varenv. *)
let new_varenv () =
  { fieldenv = StringTbl.create 100
  ; varnameenv = StringTbl.create 100
  ; typenameenv = StringTbl.create 100 }

(* Store the program's varenv and counters here. *)
let prog_var_env = ref (new_varenv ())
let prog_var_counter = ref 2
let prog_type_counter = ref 0
let prog_field_counter = ref 0

(* Operations on varenv. *)
let map_var (varstr : string) : int =
  match varstr with
      (* TODO: I think these are unnecessary. *)
    | "result" -> 0
    | "context" -> 1
    | _ ->
        let cur_var = !prog_var_counter in
          StringTbl.add !prog_var_env.varnameenv varstr cur_var;
          prog_var_counter := !prog_var_counter + 1;
          debug_print 3 ("Mapping " ^ varstr ^ " to " ^ string_of_int cur_var);
          cur_var
(* Raises Not_found if binding doesn't exist. *)
(* TODO: Decide where to handle this. *)
let lookup_var (varstr: string) : int =
  try
    match varstr with
      | "_" -> raise Not_found
      | _ -> StringTbl.find !prog_var_env.varnameenv varstr
  with _ -> raise Not_found
let unmap_var (varstr: string) : unit =
  StringTbl.remove !prog_var_env.varnameenv varstr

let map_type (varstr : string) : int =
  let cur_var = !prog_var_counter in
  StringTbl.add !prog_var_env.typenameenv varstr cur_var;
  prog_var_counter := !prog_var_counter + 1;
  cur_var
let lookup_type (varstr : string) : int =
  StringTbl.find !prog_var_env.typenameenv varstr
  
let reset_field_counter () = prog_field_counter := 0
let map_field (varstr : string) : int =
  let cur_var = !prog_field_counter in
    StringTbl.add !prog_var_env.varnameenv varstr cur_var;
    prog_field_counter := !prog_field_counter + 1;
    cur_var

