(* Header section *)
{
  open Printf
  open Yparser
  open Lexing
  let incr_lineno lexbuf =
    let pos = lexbuf.lex_curr_p in
    lexbuf.lex_curr_p <- { pos with
      pos_lnum = pos.pos_lnum + 1;
      pos_bol = pos.pos_cnum;
    }
}

(*
    Definitions section
    Simplify scanner specification with expressions of the form
      let ident = regexp
*)
let digit = ['0'-'9']
let id = ['a'-'z' '_'] ['a'-'z' '0'-'9' '_' '-' '$']*

rule token = parse
  (* Arithmetic *)
    "sat"   { SAT }
  | "unsat" { UNSAT }
  | digit digit* as num
    { INTEGER (int_of_string num) }
  | '+'   { PLUS }
  | '-'   { MINUS }
  | '*'   { MULTIPLY }
  | '/'   { DIVIDE }
  | '^'   { CARET }
  | id as ident { IDENTIFIER ident }
  (* Whitespace *)
  | [' ' '\t']  { token lexbuf }
  | '\n'  { incr_lineno lexbuf;
            NEWLINE }
  (* Equality *)
  | "=" { EQUALS }
  | "!=" { NEQUALS }
  | "<" { LT }
  | "<=" { LTE }
  | ">" { GT }
  | ">=" { GTE }
  | "&&" { AND }
  | "||" { OR }
  (* Enclosing symbols *)
  | "(" { LPAREN }
  | ")" { RPAREN }
  | _ { token lexbuf }
  | eof   { EOF }

(* Trailer section *)
{

}

