/* file: parser.mly */
/* SketchPADS parser */

/* Header section */
%{
open Lexing
open Error
%}

%token SAT UNSAT
%token NEWLINE COMMENT EOF
%token LPAREN RPAREN
%token EQUALS NEQUALS
/* Boolean */
%token LT LTE GT GTE
%token AND OR
/* Arithmetic */
%token PLUS MINUS MULTIPLY DIVIDE NEG CARET
/* Keywords */
/* Basics */
%token <int> INTEGER
%token <string> IDENTIFIER

%left PLUS MINUS
%left MULTIPLY DIVIDE
%left NEG   /* negation -- unary minus */
%right CONS
%right CARET  /* exponentiation */

%start input
%type <Yices.yexp list> input

/* Grammar follows */
%%
input: SAT NEWLINE satexps optionalvwsp optionaleof
        { $3 }
     | UNSAT { raise Unsat }
optionaleof: /* Nothing */ { }
          | EOF { } 

satexps: /* Nothing */ { [] }
       | satexp NEWLINE { [$1] }
       | satexps satexp optionalvwsp { List.append $1 [$2] }
;
vwhitespace: /* empty */ { }
           | NEWLINE  { }
           | vwhitespace NEWLINE { }
optionalvwsp: /* empty */ { }
            | vwhitespace { }

satexp: LPAREN EQUALS IDENTIFIER number RPAREN
          { Yices.YEqual ($3, $4) }
      | LPAREN EQUALS IDENTIFIER MINUS number RPAREN
          { Yices.YEqual ($3, -$5) }
;
number : INTEGER { $1 }
      | INTEGER DIVIDE INTEGER
        { let f1 = float_of_int $1 in
          let f2 = float_of_int $3 in
            truncate (f1 /. f2) }
